import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import logging
import sys
import yaml
import copy
import os
import gzip
import threading
import hashlib
import re
import pydantic
import typing
from redvypr.device import RedvyprDevice
from redvypr.data_packets import check_for_command
#from redvypr.redvypr_packet_statistic import do_data_statistics, create_data_statistic_dict

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('rawdatareplay')
logger.setLevel(logging.DEBUG)

class DeviceBaseConfig(pydantic.BaseModel):
    publishes: bool = True
    subscribes: bool = False
    description: str = "Replays a raw redvypr data file"
    gui_tablabel_display: str = 'Replay status'
    gui_icon: str = 'mdi.code-json'

class DeviceCustomConfig(pydantic.BaseModel):
    files: list = pydantic.Field(default=[], description='List of files to replay')
    replay_index: list = pydantic.Field(default=['0,-1,1'], description='The index of the packets to be replayed [start, end, nth]')
    loop: bool = pydantic.Field(default=False, description='Loop over all files if set')
    speedup: float = pydantic.Field(default=1.0, description='Speedup factor of the data')
    replace_time: bool = pydantic.Field(default=False, description='Replaces the original time in the packet with the time the packet was read')


redvypr_devicemodule = True

def get_packets(filestream=None):
    funcname = __name__ + '.get_packets()'
    packets = []
    if(filestream is not None):
        data = filestream.read()
        for databs in data.split('...\n'): # Split the text into single subpackets
            try:
                data = yaml.safe_load(databs)
                if(data is not None):
                    packets.append(data)
                    
            except Exception as e:
                logger.debug(funcname + ': Could not decode message {:s}'.format(str(databs)))
                data = None
    
        return packets
    else:
        return None


def index_file(filestream, chunksize, statusqueue=None):
    funcname = __name__ + '.index_file()'
    npackets_read = 0
    packets = []
    status_thread = {}
    tstatus = time.time()
    stat = {}
    stat['packets_seek'] = []
    stat['packets_size'] = []
    stat['packets_num'] = []
    stat['packets_t'] = []
    stat['npackets'] = 0
    packets_ = []
    seek_start = 0
    seek_now = 0
    nread = 0
    data = ''
    data_buffer = b''
    flag_eof = False
    filestream.seek(0)
    if (filestream is not None):
        while True:
            seek_start = filestream.tell()
            data_read = filestream.read(chunksize)
            seek_now = filestream.tell()
            #print('data read', seek_start,seek_now, seek_now - seek_start,len(data_read))
            lendata = len(data_read)
            nread += lendata
            # print('len', len(data_read))
            if len(data_read) < chunksize:
                flag_eof = True

            # Add potentially old data and the newly read data
            data_buffer += data_read
            seek_data_buffer_end = seek_now
            seek_data_buffer_start = seek_data_buffer_end - len(data_buffer)
            while True:
                # Look for the start of a packet
                try:
                    index_start = data_buffer.index(b'---')
                except Exception as e:
                    index_start = None

                # Look for the end of a packet
                try:
                    pattern_end = b'\0'
                    index_end = data_buffer.index(pattern_end)
                    #index_end += len(pattern_end)
                except Exception as e:
                    index_end = None

                #print('data',data_buffer)
                #print('index start',index_start,index_end)

                if (index_end is not None) and (index_start is not None) and ((index_end - index_start) > 0):
                    datab = data_buffer[index_start:index_end]
                    databs = datab.decode('utf-8')
                    #print('databs',databs,index_start,index_end)
                else:
                    break
                try:
                    data_packet = yaml.safe_load(databs)
                    if (data_packet is not None):
                        numpacket = data_packet['_redvypr']['numpacket']
                        tpacket = data_packet['_redvypr']['t']
                        #print('Found datapacket',seek_data_buffer_start)
                        loc_packet_start = seek_data_buffer_start + index_start
                        loc_packet_length = index_end - index_start
                        packets.append(data_packet)
                        stat['packets_seek'].append(loc_packet_start)
                        stat['packets_size'].append(loc_packet_length)
                        stat['packets_num'].append(numpacket)
                        stat['packets_t'].append(tpacket)

                        npackets_read += 1
                        #print('fdsfd',npackets_read)
                        stat['npackets'] = npackets_read

                        if True:
                            dt = time.time() - tstatus
                            if dt > 0.5:
                                tstatus = time.time()
                                tmin = stat['packets_t'][0]
                                tmax = stat['packets_t'][-1]
                                status_thread['t'] = time.time()
                                status_thread['t_min'] = tmin
                                status_thread['t_max'] = tmax
                                status_thread['seek'] = seek_now
                                status_thread['packets_num'] = npackets_read
                                status_thread['flag_eof'] = flag_eof
                                status_thread['stat'] = None
                                logger.debug(funcname + ' Status:' + str(status_thread))
                                if statusqueue is not None:
                                    statusqueue.put(status_thread)
                        # Remove the packet from the dta_buffer
                        data_buffer = data_buffer[index_end+len(pattern_end):]
                        seek_data_buffer_start = seek_data_buffer_end - len(data_buffer)
                except Exception as e:
                    logger.debug(funcname + ': Could not decode message {:s}'.format(str(databs)))
                    logger.exception(e)
                    #return [packets, packet_ind]

                #break


            if flag_eof:  # EOF, cleanup
                logger.debug(funcname + ': EOF. Rewinding file')
                filestream.seek(0)
                t = time.time()
                td = datetime.datetime.fromtimestamp(t)
                tdstr = td.strftime("%Y-%m-%d %H:%M:%S.%f")
                break

        # In thread mode, add stat to status dictionary
        if statusqueue is not None:
            status_thread['t'] = time.time()
            status_thread['seek'] = seek_now
            status_thread['packets_num'] = npackets_read
            status_thread['flag_eof'] = flag_eof
            status_thread['stat'] = stat
            statusqueue.put(status_thread)

        return stat


class packetreader():
    def __init__(self, filename=None, replay_index = '0,-1,1', npackets = 10, chunksize=1024,statusqueue=None,filestat = None):
        funcname = self.__class__.__name__ + '.__init__()'
        self.filename = filename
        #
        self.ipacket = -1
        self.packet_index = []
        self.npackets_read = 0
        self.nread = 0 # Amount of data read
        self.statusqueue = statusqueue
        #
        t = time.time()
        td = datetime.datetime.fromtimestamp(t)
        tdstr = td.strftime("%Y-%m-%d %H:%M:%S.%f")
        sstr = '{:s}: Opening file {:s}'.format(tdstr, filename)
        try:
            statusqueue.put_nowait(sstr)
        except:
            pass
        logger.info(sstr)
        if filename.lower().endswith('.gz'):
            FLAG_GZIP = True
        else:
            FLAG_GZIP = False

        if FLAG_GZIP:
            try:
                filestream = gzip.open(filename, 'rb')
                logger.debug(funcname + ' Opened file: {:s}'.format(filename))
            except Exception as e:
                logger.warning(funcname + ' Error opening file:' + filename + ':' + str(e))
                return None
        else:
            try:
                filestream = open(filename,'rb')
                logger.debug(funcname + ' Opened file: {:s}'.format(filename))
            except Exception as e:
                logger.warning(funcname + ' Error opening file:' + filename + ':' + str(e))
                return None

        self.filestream = filestream
        # Get the size of the data (within the file, this is different to the filesize, which can be gzipped
        self.fsize = os.path.getsize(filename)
        f = self.filestream
        f.seek(0, os.SEEK_END)
        size = f.tell() # The size of the internal data
        self.datasize = size
        self.filestream.seek(0)
        self.chunksize = chunksize

        # Calculate hash
        #hasher = hashlib.sha256()
        hasher = hashlib.md5()
        while True:
            data = f.read(65536)
            if not data:
                break
            hasher.update(data)
        self.checksum = hasher.hexdigest()
        self.filestream.seek(0)
        #print("MD5 checksum of {}".format(self.checksum))

        if filestat is None:
            # Check for an index file
            filename_index = filename + '.index{}.yaml.gz'.format(self.checksum)
            try:
                logger.debug(funcname + ' loading index file')
                filestream_index = gzip.open(filename_index, 'rb')
                filestat_raw = filestream_index.read()
                filestat = yaml.safe_load(filestat_raw)
                filestat['checksum'] = self.checksum
                logger.debug(funcname + ' loading index file done')
            except:
                logger.debug(funcname, exc_info=True)
                logger.info("Did not find index file, creating one")
                filestat = self.index_file()
                findex = gzip.open(filename_index, 'wb')
                yamlstr = yaml.dump(filestat)
                findex.write(yamlstr.encode('utf-8'))
                findex.close()

            # Inspect file, if not done already
            logger.debug(funcname + ' Indexing file {:s}'.format(filename))
            self.filestat = filestat
        elif filestat == 'thread':
            logger.debug(funcname + ' Starting thread based file statistic')
            self.index_file_thread()
        else:
            self.filestat = filestat

    def index_to_packetindex(self, indexstr):
        """

        """
        rs = indexstr.split(',')
        istart = int(rs[0])
        iend = int(rs[1])
        istep = int(rs[2])
        npackets = self.filestat['npackets']
        if iend < 0:
            iend_abs = npackets + iend + 1
        else:
            iend_abs = iend

        if istart < 0:
            istart_abs = npackets + istart
        else:
            istart_abs = istart


        packetindex = []
        for i in range(istart_abs,iend_abs,istep):
            packetindex.append(i)

        #print('Packetindex',packetindex)
        return packetindex

    def get_packets(self, indexstr):
        packetindex = self.index_to_packetindex(indexstr)
        packets = self.get_packets_by_index(packetindex)
        return packets

    def get_packets_by_index(self, packetindex):
        """
        Get the packets
        """
        packets = []
        for pindex in packetindex:
            iseek = self.filestat['packets_seek'][pindex]
            plen  = self.filestat['packets_size'][pindex]
            self.filestream.seek(iseek)
            packetdata_raw = self.filestream.read(plen)
            packetdata = yaml.safe_load(packetdata_raw)
            packets.append(packetdata)
            #print('Packet',pindex)
            #print('Packetdata',packetdata)

        return packets


    def close_file(self):
        self.filestream.close()

    def index_file(self):
        stat = index_file(self.filestream,self.chunksize)
        return stat

    def index_file_thread(self):
        self.stat_thread = {}
        self.statusqueue = queue.Queue()
        self.index_thread = threading.Thread(target = index_file, args = (self.filestream,self.chunksize,self.statusqueue), daemon = True)
        self.index_thread.start()




def packet_read_thread(filename, chunksize, npacket_buf=10, dataqueue=None, commandqueue=None, statusqueue=None):
    funcname = __name__ + '.packet_read_thread()'

    if filename.lower().endswith('.gz'):
        FLAG_GZIP = True
    else:
        FLAG_GZIP = False

    if FLAG_GZIP:
        try:
            filestream = gzip.open(filename, 'rb')
            logger.debug(funcname + ' Opened file: {:s}'.format(filename))
        except Exception as e:
            logger.warning(funcname + ' Error opening file:' + filename + ':' + str(e))
            return None
    else:
        try:
            filestream = open(filename, 'rb')
            logger.debug(funcname + ' Opened file: {:s}'.format(filename))
        except Exception as e:
            logger.warning(funcname + ' Error opening file:' + filename + ':' + str(e))
            return None

    # Get the size of the data (within the file, this is different to the filesize, which can be gzipped
    fsize = os.path.getsize(filename)
    f = filestream
    f.seek(0, os.SEEK_END)
    size = f.tell()
    datasize = size
    filestream.seek(0)

    filename_base = os.path.basename(filename)
    filename_path = os.path.dirname(filename)

    npackets_read = 0
    packets = []
    status_thread = {}
    tstatus = time.time()
    stat = {}
    stat['packets_seek'] = []
    stat['packets_size'] = []
    stat['packets_num'] = []
    stat['packets_t'] = []
    stat['npackets'] = 0
    packets_ = []
    seek_start = 0
    seek_now = 0
    nread = 0
    data = ''
    data_buffer = b''
    flag_eof = False
    filestream.seek(0)
    nread = 0
    nnewread = 0
    if (filestream is not None):
        while True:
            com = commandqueue.get()
            if type(com) == int: # Read n new packets
                nnewread = com
                nread = 0
            elif com == 'stop':  # Read n new packets
                logger.debug(funcname + ' Stopping ...')
                return
            while nread < nnewread:
                #print('Nread:',nread)
                seek_start = filestream.tell()
                data_read = filestream.read(chunksize)
                seek_now = filestream.tell()
                #print('data read', seek_start,seek_now, seek_now - seek_start,len(data_read))
                lendata = len(data_read)
                # print('len', len(data_read))
                if len(data_read) < chunksize:
                    flag_eof = True

                # Add potentially old data and the newly read data
                data_buffer += data_read
                seek_data_buffer_end = seek_now
                seek_data_buffer_start = seek_data_buffer_end - len(data_buffer)
                while True:
                    # Look for the start of a packet
                    try:
                        index_start = data_buffer.index(b'---')
                    except Exception as e:
                        index_start = None

                    # Look for the end of a packet
                    try:
                        pattern_end = b'\0'
                        index_end = data_buffer.index(pattern_end)
                        #index_end += len(pattern_end)
                    except Exception as e:
                        index_end = None

                    #print('data',data_buffer)
                    #print('index start',index_start,index_end)

                    if (index_end is not None) and (index_start is not None) and ((index_end - index_start) > 0):
                        datab = data_buffer[index_start:index_end]
                        databs = datab.decode('utf-8')
                        #print('databs',databs,index_start,index_end)
                    else:
                        break
                    try:
                        data_packet = yaml.safe_load(databs)
                        if (data_packet is not None):
                            numpacket = data_packet['_redvypr']['numpacket']
                            tpacket = data_packet['_redvypr']['t']
                            #print('Found datapacket',seek_data_buffer_start)
                            loc_packet_start = seek_data_buffer_start + index_start
                            loc_packet_length = index_end - index_start
                            #packets.append(data_packet)
                            dataqueue.put(data_packet)
                            nread += 1
                            #stat['packets_seek'].append(loc_packet_start)
                            #stat['packets_size'].append(loc_packet_length)
                            #stat['packets_num'].append(numpacket)
                            #stat['packets_t'].append(tpacket)

                            npackets_read += 1
                            #print('fdsfd',npackets_read)
                            #stat['npackets'] = npackets_read

                            if True:
                                dt = time.time() - tstatus
                                if dt > 0.5:
                                    tstatus = time.time()
                                    #logger.debug(funcname + ' Status:' + str(status_thread))
                            # Remove the packet from the dta_buffer
                            data_buffer = data_buffer[index_end+len(pattern_end):]
                            seek_data_buffer_start = seek_data_buffer_end - len(data_buffer)
                    except Exception as e:
                        logger.debug(funcname + ': Could not decode message:"{:s}"'.format(str(databs)))

                        logger.exception(e)
                        #return [packets, packet_ind]

                    #break


                if flag_eof:  # EOF, cleanup
                    logger.debug(funcname + ': EOF. Rewinding file')
                    filestream.seek(0)
                    t = time.time()
                    td = datetime.datetime.fromtimestamp(t)
                    tdstr = td.strftime("%Y-%m-%d %H:%M:%S.%f")
                    return

            # In thread mode, add stat to status dictionary
            if statusqueue is not None:
                ['time', 'filename', 'seek', 'fsize', 'pc', 'packets_num']
                status_thread['t'] = time.time()
                td = datetime.datetime.fromtimestamp(status_thread['t'])
                status_thread['time'] = td.strftime('%d %b %Y %H:%M:%S')
                status_thread['filename'] = filename_base
                status_thread['filepath'] = filename_path
                status_thread['seek'] = seek_now
                status_thread['datasize'] = size
                status_thread['filesize'] = fsize
                try:
                    pc = seek_now / size * 100
                except:
                    pc = 'NaN'
                status_thread['pc'] = "{:.2f}".format(pc)
                status_thread['packets_num'] = npackets_read
                status_thread['flag_eof'] = flag_eof
                #status_thread['stat'] = stat
                statusqueue.put(status_thread)






def start(device_info, config={'filename': ''}, dataqueue=None, datainqueue=None, statusqueue=None):
    funcname = __name__ + '.start()'
    logger.debug(funcname + ':Opening reading:')

    files = list(config['files'])
    replay_index = list(config['replay_index'])
    t_status = time.time()
    #dt_status = 2 # Status update
    dt_status = .5  # Status update
    t_sent = 0 # The time the last packets was sent
    t_packet_old = 1e12 # The time the last packet had (internally)
    try:
        config['speedup']
    except:
        config['speedup'] = 1.0 # Realtime
    
    speedup = config['speedup']    
    #
    try:
        config['loop']
    except:
        config['loop'] = False
        
    loop = config['loop']
    
    
    #statistics = create_data_statistic_dict()
    
    bytes_read         = 0
    packets_published  = 0
    dt_packet_sum      = 0
    bytes_read_total   = 0
    packets_read_total = 0
    
    tfile           = time.time() # Save the time the file was created
    tflush          = time.time() # Save the time the file was created
    FLAG_NEW_FILE = True
    nfile = 0
    packets = []
    read_dataqueue = queue.Queue()
    read_commandqueue = queue.Queue()
    while True:
        tcheck = time.time()
        try:
            data = datainqueue.get(block=False)
        except:
            data = None
        if (data is not None):
            command = check_for_command(data, thread_uuid=device_info['thread_uuid'])
            logger.debug('Got a command: {:s}'.format(str(data)))
            if (command is not None):
                sstr = funcname + ': Command is for me: {:s}'.format(str(command))
                logger.debug(sstr)
                if command == 'stop':
                    logger.debug('Stopping')
                    try:
                        statusqueue.put_nowait(sstr)
                    except:
                        pass

                    try:
                        read_commandqueue.put('stop')
                        logger.debug('stopping read thread')
                    except:
                        logger.debug('stopping read thread failed:',exc_info=True)
                    break

        if (FLAG_NEW_FILE):
            if (nfile >= len(files)):
                if (loop == False):
                    sstr = funcname + ': All files read, stopping now.'
                    try:
                        statusqueue.put_nowait(sstr)
                    except:
                        pass
                    logger.info(sstr)
                    break
                else:
                    sstr = funcname + ': All files read, loop again.'
                    try:
                        statusqueue.put_nowait(sstr)
                    except:
                        pass

                    logger.info(sstr)
                    nfile = 0

            filename = files[nfile]
            chunksize = 5000
            npacket_buf = 10
            nfile += 1
            print('Starting reading thread')
            args = (filename, chunksize, npacket_buf, read_dataqueue, read_commandqueue, statusqueue)
            read_thread = threading.Thread(target=packet_read_thread, args=args, daemon=True)
            read_thread.start()
            read_commandqueue.put(npacket_buf)
            for i in range(npacket_buf):
                packets.append(read_dataqueue.get())

            pnow = packets.pop(0)
            pnext = packets.pop(0)
            FLAG_NEW_FILE = False

        # Check if the read thread is still alive
        if not(read_thread.is_alive()):
            logger.debug(funcname + ' Reading thread finished')
            FLAG_NEW_FILE = True
        else:
            if len(packets) > 1:
                while True:
                    try:
                        packets.append(read_dataqueue.get_nowait())
                    except:
                        break
                if True:
                    t_pnow = pnow['_redvypr']['t']
                    t_pnext = pnext['_redvypr']['t']
                    dt = t_pnext - t_pnow
                    t_now = time.time()
                    if config['replace_time']:
                        pnow['t'] = t_now
                        pnow['_redvypr']['t'] = t_now

                    #print('sending',pnow)
                    dataqueue.put(pnow)
                    pnow = pnext
                    pnext = packets.pop(0)
                    packets_published += 1
                    dt_packet = dt / speedup
                    dt_packet_sum += dt_packet


                    if (dt_packet < 0):
                        dt_packet = 0
                    if (dt_packet > 10):
                        logger.warning(funcname + ' Long dt_packet of {:f} seconds'.format(dt_packet))


                    if len(packets) < npacket_buf:
                        dn = npacket_buf - len(packets)
                        #print('Asking for new packets',dn)
                        read_commandqueue.put(dn)

                    #print('sleeping dt_packet',dt_packet)
                    time.sleep(dt_packet)
                    t_sent = time.time()


        # Status update
        if (time.time() - t_status) > dt_status:
            #print('status')
            t_status = time.time()
            td = datetime.datetime.fromtimestamp(t_status)
            tdstr = td.strftime("%Y-%m-%d %H:%M:%S.%f")
            try:
                dt_avg = dt_packet_sum / packets_published
            except:
                dt_avg = -1
            sstr = '{:s}: Sent {:d} packets with an avg dt of {:.3f}s.'.format(tdstr, packets_published,
                                                                               dt_avg)
            logger.debug(sstr)
            try:
                statusqueue.put_nowait(sstr)
            except:
                pass



#
#
# The init widget
#
#
class initDeviceWidget(QtWidgets.QWidget):
    connect      = QtCore.pyqtSignal(RedvyprDevice) # Signal requesting a connect of the datainqueue with available dataoutqueues of other devices
    def __init__(self,device=None):
        super(QtWidgets.QWidget, self).__init__()
        layout        = QtWidgets.QGridLayout(self)
        self.file_statistics = {}
        self.device   = device
        self.label    = QtWidgets.QLabel("rawdatareplay setup")
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setStyleSheet(''' font-size: 24px; font: bold''')
        self.config_widgets= [] # A list of all widgets that can only be used of the device is not started yet
        # Input output widget
        self.inlabel  = QtWidgets.QLabel("Filenames") 
        self.inlabel.setStyleSheet(''' font-size: 20px; font: bold''')
        self.inlist   = QtWidgets.QTableWidget()

        self.inlist.setRowCount(1)
        #self.inlist.setSortingEnabled(True)
        self.col_replaypackets = 0
        self.col_npackets = 1
        self.col_tmin = 2
        self.col_tmax = 3
        self.col_fname = 4
        self.ncols = 5
        self.inlist.setColumnCount(self.ncols)
        self.__filelistheader__ = [[]] * self.ncols
        self.__filelistheader__[self.col_tmax] = 'Last date'
        self.__filelistheader__[self.col_tmin] = ' First date'
        self.__filelistheader__[self.col_npackets] = 'Packets'
        self.__filelistheader__[self.col_replaypackets] = 'Replay Packets'
        self.__filelistheader__[self.col_fname] = 'Filename'
        self.inlist.setHorizontalHeaderLabels(self.__filelistheader__)
        

        self.addfilesbtn   = QtWidgets.QPushButton("Add files")
        self.addfilesbtn.clicked.connect(self.add_files)
        self.remfilesbtn   = QtWidgets.QPushButton("Rem files")
        self.remfilesbtn.clicked.connect(self.rem_files)
        self.scanfilesbtn   = QtWidgets.QPushButton("Scan files")
        self.scanfilesbtn.clicked.connect(self.scan_files_clicked)
        self.config_widgets.append(self.inlist)
        self.config_widgets.append(self.addfilesbtn)


        # Looping the data?
        self.loop_checkbox = QtWidgets.QCheckBox('Loop')
        loopflag = bool(self.device.custom_config.loop)
        self.loop_checkbox.setChecked(loopflag)
        self.loop_checkbox.stateChanged.connect(self.speedup_changed)
        self.replace_time_checkbox = QtWidgets.QCheckBox('Replace time')
        replace_time_flag = bool(self.device.custom_config.replace_time)
        self.replace_time_checkbox.setChecked(replace_time_flag)
        self.replace_time_checkbox.stateChanged.connect(self.speedup_changed)
        # Speedup
        self.speedup_edit  = QtWidgets.QLineEdit(self)
        onlyDouble = QtGui.QDoubleValidator()
        self.speedup_edit.setValidator(onlyDouble)
        self.speedup_edit.setToolTip('Speedup of the packet replay.')
        self.speedup_label = QtWidgets.QLabel("Speedup factor")
        speedup = float(self.device.custom_config.speedup)
        self.speedup_edit.setText("{:.1f}".format(speedup))
        self.speedup_edit.textChanged.connect(self.speedup_changed)
        
        
        self.startbtn = QtWidgets.QPushButton("Start replay")
        self.startbtn.clicked.connect(self.start_clicked)
        self.startbtn.setCheckable(True)
        self.startbtn.setSizePolicy(QtWidgets.QSizePolicy.Preferred,QtWidgets.QSizePolicy.Expanding)
        
        layout.addWidget(self.label,0,0,1,-1)
        
        layout.addWidget(self.addfilesbtn,1,0,1,-1)               
        layout.addWidget(self.remfilesbtn,2,0,1,-1)          
        layout.addWidget(self.scanfilesbtn,3,0,1,-1)
        layout.addWidget(self.inlabel,4,0,1,-1,QtCore.Qt.AlignCenter)         
        layout.addWidget(self.inlist,5,0,1,-1)
        layout.addWidget(self.loop_checkbox,6,0)
        layout.addWidget(self.replace_time_checkbox, 6, 1)
        layout.addWidget(self.speedup_label,6,2,1,1,QtCore.Qt.AlignRight)
        layout.addWidget(self.speedup_edit,6,3,1,1,QtCore.Qt.AlignRight)
        layout.addWidget(self.startbtn,7,0,2,-1)


        # Update the widgets depending on the configuration
        self.update_filenamelist()
        self.statustimer = QtCore.QTimer()
        self.statustimer.timeout.connect(self.update_buttons)
        self.statustimer.start(500)

    def speedup_changed(self):
        funcname = self.__class__.__name__ + '.speedup_changed()'
        logger.debug(funcname)
        # Speedup
        self.device.custom_config.speedup = float(self.speedup_edit.text())
        loopflag = self.loop_checkbox.isChecked()
        replace_time_flag = self.replace_time_checkbox.isChecked()
        self.device.custom_config.loop = loopflag
        self.device.custom_config.replace_time = replace_time_flag

    def table_changed(self,row,col):
        funcname = self.__class__.__name__ + '.table_changed()'
        logger.debug(funcname)
        #print('Row',row,'Col',col)
        item = self.inlist.item(row,col)

        if col == self.col_replaypackets: # If the replay index was changed
            rindex = item.text()
            try:
                rs = rindex.split(',')
                istart = int(rs[0])
                iend = int(rs[1])
                istep = int(rs[2])
                self.device.custom_config.replay_index[item.replay_index] = rindex
                item.replay_index_str = rindex
            except Exception as e:
                logger.exception(e)
                replay_index_str = item.replay_index_str
                itemold = QtWidgets.QTableWidgetItem(str(replay_index_str))
                itemold.replay_index_str = replay_index_str
                try:
                    self.inlist.cellChanged.disconnect(self.table_changed)
                except:
                    pass
                self.inlist.setItem(row, col, itemold)
                self.inlist.cellChanged.connect(self.table_changed)


    def update_filenamelist(self):
        """ Update the filetable
        """
        try:
            funcname = self.__class__.__name__ + '.update_filenamelist()'
            logger.debug(funcname)
            try:
                self.inlist.cellChanged.disconnect(self.table_changed)
            except:
                pass
            self.inlist.clear()
            self.inlist.setHorizontalHeaderLabels(self.__filelistheader__)
            nfiles = len(self.device.custom_config.files)
            self.inlist.setRowCount(nfiles)
            rows = []
            for i, f in enumerate(self.device.custom_config.files):
                if len(self.device.custom_config.replay_index) < (i + 1):
                    self.device.custom_config.replay_index.append(self.device.custom_config.replay_index[-1])
                    replayindex = self.device.custom_config.replay_index[-1]
                else:
                    replayindex = self.device.custom_config.replay_index[i]

                item = QtWidgets.QTableWidgetItem(str(replayindex))
                item.replay_index = i
                item.replay_index_str = replayindex
                self.inlist.setItem(i, self.col_replaypackets, item)
                # Filename
                item = QtWidgets.QTableWidgetItem(str(f))
                item.setFlags(item.flags() & ~QtCore.Qt.ItemIsEditable)
                self.inlist.setItem(i, self.col_fname, item)
                rows.append(i)

            self.inlist.resizeColumnsToContents()
            #for i, f in enumerate(self.device.config.files):
            #    self.scan_file(str(f), i)

            self.inlist.resizeColumnsToContents()

            # Loop flag
            loop = bool(self.device.custom_config.loop)
            self.loop_checkbox.setChecked(loop)
            speedupstr = "{:.1f}".format(float(self.device.custom_config.speedup))
            self.speedup_edit.setText(speedupstr)
            ## Add the packetnumber etc etc
            #self.scan_files(rows)
            self.inlist.cellChanged.connect(self.table_changed)
        except Exception as e:
            logger.exception(e)








    def finalize_init(self):
        """ Util function that is called by redvypr after initializing all config (i.e. the configuration from a yaml file)
        """
        funcname = self.__class__.__name__ + '.finalize_init()'
        logger.debug(funcname)
        
    def scan_files_clicked(self):
        """ Scans the selected files from the files list for possible datastreams 
        """
        funcname = self.__class__.__name__ + '.scan_files_clicked()'
        logger.debug(funcname)
        rows = []
        for i in self.inlist.selectionModel().selection().indexes():
            row, column = i.row(), i.column()
            rows.append(row)
    
        self.scan_files(rows)

    def scan_file(self, filename, row):
        """ Scans the selected files from the files list for possible datastreams
        """
        funcname = self.__class__.__name__ + '.scan_files()'
        logger.debug(funcname)

        stat = self.inspect_data(filename,rescan=False)
        if True:
            npackets = len(stat['packets_num'])
            t_min = stat['packets_t'][0]
            t_max = stat['packets_t'][-1]
            packetitem = QtWidgets.QTableWidgetItem(str(npackets))
            packetitem.setFlags(packetitem.flags() & ~QtCore.Qt.ItemIsEditable)
            self.inlist.setItem(row, self.col_npackets, packetitem)
            tdmin = datetime.datetime.fromtimestamp(t_min)
            tminstr = str(tdmin)
            tdmax = datetime.datetime.fromtimestamp(t_max)
            tmaxstr = str(tdmax)
            t_min_item = QtWidgets.QTableWidgetItem(tminstr)
            t_min_item.setFlags(t_min_item.flags() & ~QtCore.Qt.ItemIsEditable)
            t_max_item = QtWidgets.QTableWidgetItem(tmaxstr)
            t_max_item.setFlags(t_max_item.flags() & ~QtCore.Qt.ItemIsEditable)
            self.inlist.setItem(row, self.col_tmin, t_min_item)
            self.inlist.setItem(row, self.col_tmax, t_max_item)

        self.inlist.resizeColumnsToContents()

    def inspect_data(self, filename, rescan=False):
        """ Inspects the files for possible datastreams in the file with the filename located in config['files'][fileindex].
        """
        funcname = self.__class__.__name__ + '.inspect_data()'
        logger.debug(funcname)

        try:
            stat = self.file_statistics[filename]
            FLAG_HASSTAT = True
        except:
            FLAG_HASSTAT = False

        if (rescan or (FLAG_HASSTAT == False)):
            logger.debug(funcname + ': Scanning file {:s}'.format(filename))
            p = packetreader(filename)
            # stat = p.ipacketreadernspect_file()
            self.file_statistics[filename] = p.filestat
        else:
            logger.debug(funcname + ': No rescan of {:s}'.format(filename))

        return self.file_statistics[filename]
        
    def scan_files(self,rows):
        """ Scans the selected files from the files list for possible datastreams 
        """
        funcname = self.__class__.__name__ + '.scan_files()'
        logger.debug(funcname)

        for i in rows:
            filename = self.inlist.item(i,self.col_fname).text()
            #stat = self.inspect_data_thread(filename, i, rescan=False)

        # Start a timer to update
        #if len(rows) > 0:
        #    self.threadtimer = QtCore.QTimer()
        #    self.threadtimer.timeout.connect(self.update_table_from_thread)  # Add to the timer another update
        #    self.threadtimer.start(250)
            
        self.inlist.resizeColumnsToContents()

    def inspect_data_thread(self, filename, row, rescan=False):
        """ Inspects the files for possible datastreams in the file with the filename located in config['files'][fileindex].
        """
        funcname = self.__class__.__name__ + '.inspect_data()'
        logger.debug(funcname)
        self.inspect_threads = []

        try:
            stat = self.file_statistics[filename]
            FLAG_HASSTAT = True
        except:
            FLAG_HASSTAT = False

        if (rescan or (FLAG_HASSTAT == False)):
            logger.debug(funcname + ': Scanning file {:s}'.format(filename))
            p = packetreader(filename, filestat='thread')
            p.row = row
            self.inspect_threads.append(p)
            # stat = p.inspect_file()
            # self.file_statistics[filename] = p.filestat
        else:
            logger.debug(funcname + ': No rescan of {:s}'.format(filename))

    def update_table_from_thread(self):
        funcname = self.__class__.__name__ + '.update_table_from_thread()'
        logger.debug(funcname)
        flag_continue = False
        for p in self.inspect_threads:
            try:
                status = p.statusqueue.get_nowait()
                print(funcname + ' Got status')
                # print('Hallo status',status)
                row = p.row
            except Exception as e:
                status = None
                pass
            if p.index_thread.is_alive() == True:
                flag_continue = True

            if status is not None:
                packetitem = QtWidgets.QTableWidgetItem(str(status['packets_num']))
                packetitem.setFlags(packetitem.flags() & ~QtCore.Qt.ItemIsEditable)
                self.inlist.setItem(row, self.col_npackets, packetitem)
                tdmin = datetime.datetime.fromtimestamp(status['t_min'])
                tminstr = str(tdmin)
                tdmax = datetime.datetime.fromtimestamp(status['t_max'])
                tmaxstr = str(tdmax)
                t_min_item = QtWidgets.QTableWidgetItem(tminstr)
                t_min_item.setFlags(t_min_item.flags() & ~QtCore.Qt.ItemIsEditable)
                t_max_item = QtWidgets.QTableWidgetItem(tmaxstr)
                t_max_item.setFlags(t_max_item.flags() & ~QtCore.Qt.ItemIsEditable)
                self.inlist.setItem(row, self.col_tmin, t_min_item)
                self.inlist.setItem(row, self.col_tmax, t_max_item)
                self.inlist.resizeColumnsToContents()

        if flag_continue == False:
            logger.debug('No thread running anymore')
            self.threadtimer.stop()
            
    def rem_files(self):
        """ Remove the selected files from the files list
        """
        funcname = self.__class__.__name__ + '.rem_files()'
        logger.debug(funcname)
        rows = []
        for i in self.inlist.selectionModel().selection().indexes():
            row, column = i.row(), i.column()
            rows.append(row)
        
        rows = list(set(rows))
        rows.sort(reverse=True)  
        for i in rows:
            self.device.custom_config.files.pop(i)
            
        self.update_filenamelist() 

    def add_files(self):
        """ Opens a dialog to choose file to add
        """
        funcname = self.__class__.__name__ + '.add_files()'
        regex_indexfile = re.compile('.*[.index][0-9a-f]{32}.yaml.gz')
        logger.debug(funcname)
        filenames, _ = QtWidgets.QFileDialog.getOpenFileNames(self,"Rawdatafiles","","redvypr raw gzip (*.redvypr_yaml.gz);;redvypr raw (*.redvypr_yaml);;All Files (*)")
        for f in filenames:
            if regex_indexfile.match(f) is None:
                self.device.custom_config.files.append(f)
            else:
                logger.info('Found index file {}, will not use it'.format(f))
            
        self.update_filenamelist()
        self.inlist.sortItems(self.col_fname, QtCore.Qt.AscendingOrder)
            

        
    def resort_files(self):
        """ Resorts the files in config['files'] according to the sorting in the table
        """
        files_new = []
        nfiles = len(self.device.custom_config.files)
        for i in range(nfiles):
            filename = self.inlist.item(i,self.col_fname).text()
            files_new.append(filename)
            
        self.device.custom_config.files = files_new
        
    def con_clicked(self):
        funcname = self.__class__.__name__ + '.con_clicked():'
        logger.debug(funcname)
        button = self.sender()
        if(button == self.adddeviceinbtn):
            self.connect.emit(self.device)
            
    def start_clicked(self):
        funcname = self.__class__.__name__ + '.start_clicked():'
        logger.debug(funcname)
        button = self.sender()
        if button.isChecked():
            logger.debug(funcname + "button pressed")
            # Update the config for replay
            # update loop, replace_time and speedup
            self.speedup_changed()
            # Replay index
            self.device.thread_start()
        else:
            logger.debug(funcname + 'button released')
            self.device.thread_stop()

            
    def update_buttons(self):
            """ Updating all buttons depending on the thread status (if its alive, graying out things)
            """

            status = self.device.get_thread_status()
            thread_status = status['thread_running']

            # Running
            if(thread_status):
                self.startbtn.setText('Stop')
                self.startbtn.setChecked(True)
                for w in self.config_widgets:
                    w.setEnabled(False)
            # Not running
            else:
                self.startbtn.setText('Start')
                for w in self.config_widgets:
                    w.setEnabled(True)
                    
                # Check if an error occured and the startbutton 
                if(self.startbtn.isChecked()):
                    self.startbtn.setChecked(False)
                #self.conbtn.setEnabled(True)


class displayDeviceWidget(QtWidgets.QWidget):
    def __init__(self,device=None):
        super(QtWidgets.QWidget, self).__init__()
        layout          = QtWidgets.QVBoxLayout(self)
        hlayout         = QtWidgets.QFormLayout()
        self.device     = device
        self.text       = QtWidgets.QPlainTextEdit(self)
        self.text.setReadOnly(True)
        self.scrollchk  = QtWidgets.QCheckBox('Scroll to end')
        self.scrollchk.setChecked(True)
        self.statuslab= QtWidgets.QLabel("Status")
        self.text.setMaximumBlockCount(10000)
        # Add a table
        self.statustable = QtWidgets.QTableWidget()
        self.__statusheader__ = ['Time','Filename','Filesize','Bytes read','Bytes total','%','Packets read']
        self.statustable.setColumnCount(len(self.__statusheader__))
        self.statustable.setHorizontalHeaderLabels(self.__statusheader__)
        self.statustable.setRowCount(1)
        self.statustable.verticalHeader().setVisible(False)
        self.statustable.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)

        hlayout.addRow(self.statuslab)
        layout.addWidget(self.statustable, 1)
        layout.addLayout(hlayout)
        layout.addWidget(self.text, 5)
        layout.addWidget(self.scrollchk, 1)
        self.statustimer = QtCore.QTimer()
        self.statustimer.timeout.connect(self.update)
        self.statustimer.start(500)

    def update(self):
        statusqueue = self.device.statusqueue
        while (statusqueue.empty() == False):
            try:
                data = statusqueue.get(block=False)
            except:
                break

            if type(data) == dict:
                #print('data',data)
                statuskeys = ['time', 'filename', 'filesize', 'seek', 'datasize', 'pc', 'packets_num']
                for i,k in enumerate(statuskeys):
                    datastr = str(data[k])
                    dataitem = QtWidgets.QTableWidgetItem(datastr)
                    self.statustable.setItem(0, i, dataitem)
                    self.statustable.resizeColumnsToContents()
            else:
                #Original position of scrollbar
                pos = self.text.verticalScrollBar().value()
                self.text.moveCursor(QtGui.QTextCursor.End)
                self.text.insertPlainText(str(data) + '\n')
                if(self.scrollchk.isChecked()):
                    self.text.verticalScrollBar().setValue(self.text.verticalScrollBar().maximum())
                else:
                    self.text.verticalScrollBar().setValue(pos)


import sys
import logging
import copy
from .redvypr_address import RedvyprAddress
import redvypr.data_packets as data_packets
import time
import json
import deepdiff

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('redvypr_packet_statistics')
logger.setLevel(logging.INFO)

# A dictionary for the device_redvypr entry in the statistics
device_redvypr_statdict = {'_redvypr': {},'datakeys':[],'datakeys_expanded': {},'packets_received':0,'packets_published':0,'packets_droped':0,'_metadata':{},'_deviceinfo':{},'_keyinfo':{}}



def treat_datadict(data, devicename, hostinfo, numpacket, tpacket, devicemodulename=''):
    """ Treats a datadict received from a device and adds additional information from redvypr as hostinfo, numpackets etc.
    """
    # Add deviceinformation to the data package
    if ('_redvypr' not in data.keys()):
        data['_redvypr'] = {}
    if ('tag' not in data['_redvypr'].keys()): # A tag of the uuid, counting the number of times the packet has been recirculated
        data['_redvypr']['tag'] = {}
    if ('device' not in data['_redvypr'].keys()):
        data['_redvypr']['device'] = str(devicename)
    #if ('publisher' not in data['_redvypr'].keys()):
    #    data['_redvypr']['publisher'] = str(devicename)
    if ('packetid' not in data['_redvypr'].keys()):
        data['_redvypr']['packetid'] = str(devicename)
    if ('host' not in data['_redvypr'].keys()):
        data['_redvypr']['host'] = hostinfo
    elif (data['_redvypr']['host']['hostname'] is None) or (data['_redvypr']['host']['uuid'] is None):
        # Invalid host, replacing with local host
        data['_redvypr']['host'] = hostinfo

    # Tag the datapacket and add the local publishing device
    try:
        data['_redvypr']['tag'][hostinfo['uuid']] += 1
    except:
        data['_redvypr']['tag'][hostinfo['uuid']] = 1
        data['_redvypr']['localhost'] = hostinfo
        #data['_redvypr']['localuuid'] = hostinfo['uuid']
        data['_redvypr']['publisher'] = str(devicename)

    # Add the time to the datadict if its not already in
    if ('t' not in data['_redvypr'].keys()):
        data['_redvypr']['t'] = tpacket

    # Check if there was data sent (len(datakeys) > 0), if yes check if time is present, if not add it
    datakeys = get_keys_from_data(data)
    if (len(datakeys) > 0) and ('t' not in datakeys):
        data['t'] = tpacket

    # Add the devicemodulename to the redvypr
    if ('devicemodulename' not in data['_redvypr'].keys()):
        data['_redvypr']['devicemodulename'] = devicemodulename

    # Add the packetnumber to the datadict
    if ('numpacket' not in data['_redvypr'].keys()):
        data['_redvypr']['numpacket'] = numpacket

    return data
def create_data_statistic_dict():
    statdict = {}
    statdict['inspect'] = True
    statdict['packets_published'] = 0
    statdict['packets_received'] = 0
    statdict['datakeys'] = []
    #statdict['datakeys_expanded'] = {}
    statdict['devicekeys'] = {}
    statdict['devices'] = []
    statdict['devices_dict'] = {}
    statdict['datastreams'] = []
    statdict['datastreams_dict'] = {}
    statdict['datastreams_info'] = {}
    statdict['hostinfos'] = {}
    # New
    statdict['datakey_info'] = {}
    statdict['datastream_redvypr'] = {}
    statdict['device_redvypr'] = {}
    statdict['host_redvypr'] = {}
    statdict['metadata'] = {}
    statdict['packets'] = {}  # Packets from subscribed devices
    return statdict


def rem_device_from_statistics(deviceaddress, statdict):
    """
    Remove a deviceaddress from the statistic
    Args:
        deviceaddress:
        statdict:

    Returns:

    """
    keys_removed = []
    for k in statdict.keys():
        try:
            statdict[k].pop(deviceaddress)
            keys_removed.append(k)
        except:
            pass

    return keys_removed


def do_data_statistics(data, statdict, address_data = None):
    """
    Fills in the statistics dictionary with the data packet information
    :param data:
    :param statdict:
    :param address_data:
    :return: statdict, status
    """
    status = {'metadata_changed':False}
    if address_data is None:
        raddr = RedvyprAddress(data)
    else:
        raddr = address_data

    uuid = raddr.uuid
    address_str = raddr.address_str

    # Create a hostinfo information
    try:
        statdict['host_redvypr'][uuid].update(data['_redvypr']['host'])
    except:
        statdict['host_redvypr'][uuid] = data['_redvypr']['host']

    # Create device_redvypr, dictionary with all devices as keys
    try:
        statdict['device_redvypr'][address_str]['packets_published'] += 1
    except:  # Does not exist yet, create the entry
        statdict['device_redvypr'][address_str] = copy.deepcopy(device_redvypr_statdict)

    # Get datakeys from datapacket
    datakeys = get_keys_from_data(data)
    # Get datakeys from info (potentially)
    try:
        datakeys_info = get_keys_from_data(data['_keyinfo'])
    except Exception as e:
        datakeys_info = []

    try:
        datakeys_new = list(set(statdict['device_redvypr'][address_str]['datakeys'] + datakeys + datakeys_info))
    except Exception as e:
        logger.exception(e)
        datakeys_new = datakeys

    statdict['device_redvypr'][address_str]['_redvypr'].update(data['_redvypr'])
    statdict['device_redvypr'][address_str]['datakeys'] = datakeys_new

    # Metadata of the datapacket
    try:
        for address_str_metadata in data['_metadata'].keys():
            metadata = data['_metadata'][address_str_metadata]
            deephash = deepdiff.DeepHash(metadata)[metadata]
            try:
                metadata_orig = statdict['metadata'][address_str_metadata]
            except:
                statdict['metadata'][address_str_metadata] = {}
                metadata_orig = statdict['metadata'][address_str_metadata]

            deephash_orig = deepdiff.DeepHash(metadata_orig)[metadata_orig]
            # Check if there is a difference, if yes, update
            if deephash != deephash_orig:
                #print('Updating metadata')
                statdict['metadata'][address_str_metadata].update(metadata)
                status['metadata_changed'] = True
            else:
                #print('Same metadata, doing nothing')
                pass
    except:
        pass

    # TODO (0.9.2++): On the long term it shall be replaced by _metadata
    try:
        statdict['device_redvypr'][address_str]['_deviceinfo'].update(data['_deviceinfo'])
    except:
        pass

    # TODO (0.9.2++): On the long term it shall be replaced by _metadata
    try:
        statdict['device_redvypr'][address_str]['_keyinfo'].update(data['_keyinfo'])
    except:
        pass

    # Deeper check, data types and expanded data types
    rdata = data_packets.Datapacket(data)
    datakeys_expanded = rdata.datakeys(expand=True)
    #print('Datakeys expanded',datakeys_expanded)
    statdict['device_redvypr'][address_str]['datakeys_expanded'].update(datakeys_expanded)

    return statdict, status



def get_keys_from_data(data):
    """
    Returns the keys of a redvypr data packet without the potentially existing standard keys:
    -'_redvypr'
    -'_redvypr_command'
    -'_deviceinfo'
    -'_keyinfo'

    Args:
        data (dict): redvypr data dictionary
    Returns:
        list with the datakeys

    """
    keys = list(data.keys())
    for key_remove in data_packets.redvypr_data_keys:
        try:
            keys.remove(key_remove)
        except:
            pass

    return keys

def get_metadata_deviceinfo_all(statistics, address, publisher_strict=True,  mode='merge'):
    """
    Gets the metadata from a deviceinfo_all dict retreived with
    redvypr.get_deviceinfo()

    :param statistics:
    :param address:
    :param publisher_strict:
    :param mode:
    :return:
    """
    funcname = __name__ + '.get_metadata_deviceinfo_all():'
    logger.debug(funcname + '{}'.format(address))
    metadata = {}
    raddress = RedvyprAddress(address)
    if publisher_strict:
        publisher_key = raddress.publisher
        #print('Publisher key',publisher_key)
        try:
            #mdata_device = statistics['metadata'][publisher_key]
            mdata_device = {'metadata': statistics['metadata'][publisher_key]}
            mdata = get_metadata(mdata_device, address, mode)
            metadata.update(mdata)
        except:
            logger.debug('Could not find publisher for address {}'.format(raddress),exc_info=True)
    else:
        print('Statistics keys',statistics.keys())
        for dev in statistics['metadata'].keys():
            mdata_device = {'metadata': statistics['metadata'][dev]}
            mdata = get_metadata(mdata_device, address, mode)
            metadata.update(mdata)

    return metadata

def get_metadata(statistics, address, mode='merge'):
    """
    Gets the metadata of the redvypr address
    :param statistics:
    :param address:
    :param mode: merge or dict
    :return:
    """

    funcname = __name__ + '.get_metadata():'
    logger.debug(funcname)
    if mode == 'merge':
        metadata_return = {}
    else:
        metadata_return = {}
    raddress = RedvyprAddress(address)

    # Sort the datakeys by the number of the datakey indices.
    # This allows to have the longest entries latest such that the most
    # specific one is overwriting a less specific one
    #https://docs.python.org/3/howto/sorting.html Decorate-Sort-Undecorate
    decorated = [(len(RedvyprAddress(astr).get_datakeyentries()),astr) for astr in statistics['metadata'].keys()]
    decorated.sort()
    metadata_keys_sorted = [astr for nentries,astr in decorated]
    #print('Metadaty_keys_sorted',metadata_keys_sorted)
    #for astr in statistics['metadata'].keys():
    for astr in metadata_keys_sorted:
        raddr = RedvyprAddress(astr)
        if raddress.datakeyeval == False:
            if raddr in raddress:
                metadata = statistics['metadata'][astr]
                #print('Metadata', metadata)
                if mode == 'merge':
                    metadata_return.update(metadata)
                else:
                    metadata_return[astr] = metadata
        # loop over all datakeys and check for a hit
        else:
            dkeys1 = raddr.get_datakeyentries()
            dkeys2 = raddress.get_datakeyentries()
            dkeys2_fit = dkeys2[:len(dkeys1)]
            datakey_construct_new = ''
            for dentry in dkeys2_fit:
                datakey_construct_new = datakey_construct_new + '[' + json.dumps(dentry) + ']'

            raddress_construct = RedvyprAddress(address, datakey=datakey_construct_new)
            #print('Constructed address {}'.format(raddress_construct))
            if raddr in raddress_construct:
                metadata = statistics['metadata'][astr]
                #print('Metadata', metadata)
                if mode == 'merge':
                    metadata_return.update(metadata)
                else:
                    metadata_return[astr] = metadata

    return metadata_return

    #for astr in statistics['device_redvypr'].keys():
    #    raddr = RedvyprAddress(astr)
    #    if '_metadata' in statistics['device_redvypr'][astr].keys():
    #        for astr2 in statistics['device_redvypr'][astr]['_metadata'].keys():
    #            raddr2 = RedvyprAddress(astr2)
    #            # Check if a eval address is present (or not), if yes, check for all components
    #            if raddress.parsed_addrstr_expand['datakeyeval'] == False:
    #                if raddr2 in raddress:
    #                    metadata.update(statistics['device_redvypr'][astr]['_metadata'][astr2])
    #            else:
    #                print('Datakeyeval')
    #                #if self.expandlevel == 0:
    #                #    datakey_construct_new = data_new_key
    #                #else:
    #                #    datakey_construct_new = datakey_construct + '[' + json.dumps(data_new_key) + ']'
    #                datakey_construct_new = '['
    #                for dentry in raddress.parsed_addrstr_expand['datakeyentries_str']:
    #                    datakey_construct_new = datakey_construct_new + '[' + json.dumps(dentry) + ']'
    #                    raddr_construct = RedvyprAddress(address, datakey=datakey_construct_new)
    #                    print('hallo',raddr_construct)
    #                    pass



    return metadata

# Legacy
def get_devicename_from_data(data, uuid=False):
    """ Returns a redvypr devicename string including the hostname, ip with the optional uuid.

    Args:
        data (dict): A redvypr data dictionary
        uuid (Optional[bool]): Default False. Returns the devicename with the uuid (True) or the hostname + IP-Address (False)

    Returns:
        str: The full devicename
    """
    if (uuid):
        devicename = data['_redvypr']['device'] + ':' + data['_redvypr']['host']['hostname'] + '@' + \
                     data['_redvypr']['host']['addr'] + '::' + data['_redvypr']['host']['uuid']
    else:
        devicename = data['_redvypr']['device'] + ':' + data['_redvypr']['host']['hostname'] + '@' + \
                     data['_redvypr']['host']['addr']
    return devicename

# legacy
def get_datastream_from_data(data, datakey, uuid=False):
    """ Returns a redvypr datastream string including the hostname, ip with the optional uuid.

    Args:
        datakey (str): The datakey,
        data (dict): A redvypr data dictionary
        uuid (Optional[bool]): Default False. Returns the devicename with the uuid (True) or the hostname + IP-Address (False)

    Returns:
        str:
            - The full datastream
            - devicename string if datakey is None or ''
            - None if datakey is not in data dict
    """
    devicestr = get_devicename_from_data(data, uuid)
    if ((datakey == None) or (datakey == '')):
        return devicestr
    else:
        if (datakey in data.keys()):
            datastream = datakey + '/' + devicestr
            return datastream
        else:
            return None

# legacy
def get_datastreams_from_data(data, uuid=False, add_dict=False):
    """ Returns all datastreams of the datapacket

    Args:
        data (dict): A redvypr data dictionary
        uuid (Optional[bool]): Default False. Returns the devicename with the uuid (True) or the hostname + IP-Address (False)
        add_dict (Optional[bool]): Default False. Returns the devicename with the uuid (True) or the hostname + IP-Address (False)

    Returns:
        list: A list of all datastreams
    """
    t = time.time()
    datastreams = []
    datastreams_dict = {}
    datakeys = get_keys_from_data(data)
    devicename = get_devicename_from_data(data, uuid=True)
    for key in datakeys:
        datastream = key + '/' + devicename
        datastreams.append(datastream)
        datastreams_dict[datastream] = {'host': data['_redvypr']['host']}

    if (add_dict):
        return [datastreams, datastreams_dict]
    else:
        return datastreams



"""

Logger that writes xlsx files

"""

import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import logging
import sys
import yaml
import copy
import gzip
import os
import xlsxwriter
import pympler.asizeof
import pydantic
import typing
from redvypr.device import RedvyprDevice
import redvypr.data_packets as data_packets
import redvypr.redvypr_address as redvypr_address
import redvypr.packet_statistic as packet_statistics
import redvypr.gui

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('xlsxwriter')
logger.setLevel(logging.INFO)

time_format = 'yyyy-mm-dd HH:MM:SS.000'

row_address = 0
row_host = 1
row_uuid = 2
row_device = 3
row_publisher = 4
row_packetid = 5
row_datakey = 6
row_dataunit = 7
row_firstdata = row_dataunit + 1

redvypr_devicemodule = True
class DeviceBaseConfig(pydantic.BaseModel):
    publishes: bool = False
    subscribes: bool = True
    description: str = "Saves subscribed devices in a xlsx file"
    clear_datainqueue_before_thread_starts: bool = True
    gui_tablabel_display: str = 'xlsx logging status'
    gui_icon: str = 'fa5s.file-excel'

class DeviceCustomConfig(pydantic.BaseModel):
    dt_sync: int = pydantic.Field(default=5,description='Time after which an open file is synced on disk')
    dt_newfile: int = pydantic.Field(default=3600,description='Time after which a new file is created')
    dt_newfile_unit: typing.Literal['none','seconds','hours','days'] = pydantic.Field(default='seconds')
    dt_update: int = pydantic.Field(default=2,description='Time after which an upate is sent to the gui')
    datakey_expansionlevel: int = pydantic.Field(default=3, description='Level of the datakey expansionlevel')
    size_newfile:int = pydantic.Field(default=500,description='Size of object in RAM after which a new file is created')
    size_newfile_unit: typing.Literal['none','bytes','kB','MB'] = pydantic.Field(default='MB')
    datafolder:str = pydantic.Field(default='./',description='Folder the data is saved to')
    fileextension:str= pydantic.Field(default='xlsx',description='File extension, if empty not used')
    fileprefix:str= pydantic.Field(default='redvypr',description='If empty not used')
    filepostfix:str= pydantic.Field(default='xlsxwriter',description='If empty not used')
    filedateformat:str= pydantic.Field(default='%Y-%m-%d_%H%M%S',description='Dateformat used in the filename, must be understood by datetime.strftime')
    filecountformat:str= pydantic.Field(default='04',description='Format of the counter. Add zero if trailing zeros are wished, followed by number of digits. 04 becomes {:04d}')

def create_logfile(config,count=0,all_worksheets={}):
    funcname = __name__ + '.create_logfile():'
    logger.debug(funcname)

    filename = ''
    if len(config['datafolder']) > 0:
        if os.path.isdir(config['datafolder']):
            filename += config['datafolder'] + os.sep
        else:
            logger.warning(funcname + ' Data folder {:s} does not exist.'.format(filename))
            return None

    if(len(config['fileprefix'])>0):
        filename += config['fileprefix']

    if (len(config['filedateformat']) > 0):
        tstr = datetime.datetime.now().strftime(config['filedateformat'])
        filename += '_' + tstr

    if (len(config['filecountformat']) > 0):
        cstr = "{:" + config['filecountformat'] +"d}"
        filename += '_' + cstr.format(count)

    if (len(config['filepostfix']) > 0):
        filename += '_' + config['filepostfix']

    if (len(config['fileextension']) > 0):
        filename += '.' + config['fileextension']

    logger.info(funcname + ' Will create a new file: {:s}'.format(filename))

    # Create a workbook and add a worksheet.
    workbook = xlsxwriter.Workbook(filename,{'in_memory': True})
    date_format = workbook.add_format({'num_format': time_format})
    header_format = workbook.add_format({'bold': True,'bg_color':'#F0F0F0'})
    text_wrap_format = workbook.add_format({'text_wrap': True})
    all_worksheets['header_format'] = header_format
    all_worksheets['text_wrap_format'] = text_wrap_format
    formats = {}
    formats['date'] = date_format

    worksheet_summary = workbook.add_worksheet('summary')
    # Write some information
    worksheet_summary.write(0, 0, 'redvypr')
    worksheet_summary.write(0, 1, 'xlsxwriter')
    redvypr_version_str = 'redvypr {}'.format(redvypr.version)
    worksheet_summary.write(1, 0, 'version')
    worksheet_summary.write(1, 1, redvypr_version_str)
    all_worksheets['summary'] = worksheet_summary
    all_worksheets['metadata'] = workbook.add_worksheet('metadata')
    all_worksheets['metadata_indices'] = {'rows': [], 'columns': []}
    return [workbook,filename,formats]

def write_metadata(workbook, all_worksheets, datakey, data, deviceinfo_all, device_info, config):
    funcname = 'write_metadata():'
    logger.debug(funcname)
    raddress_tmp = redvypr_address.RedvyprAddress(data, datakey=datakey)
    raddress_tmp_str = raddress_tmp.get_str('/h/d/i/k')
    raddress_tmp_str_full = raddress_tmp.get_fullstr()
    metadata_tmp = packet_statistics.get_metadata_deviceinfo_all(deviceinfo_all, raddress_tmp)
    header_rows = [(row_host, raddress_tmp.hostname, 'hostname')]
    header_rows.append((row_uuid, raddress_tmp.uuid, 'uuid'))
    header_rows.append((row_device, raddress_tmp.devicename, 'devicename'))
    header_rows.append((row_publisher, raddress_tmp.publisher, 'publisher'))
    header_rows.append((row_packetid, raddress_tmp.packetid, 'packetid'))
    header_rows.append((row_datakey, datakey, 'datakey'))
    for header_row in header_rows:
        colindex = 0
        all_worksheets['metadata'].write(header_row[0], colindex, header_row[1],
                                                    all_worksheets['header_format'])
        all_worksheets['metadata'].write(header_row[0], 0, header_row[2], all_worksheets['header_format'])
    #device_worksheets[packet_address_str].write(lineindex, colindex, datawrite)
    if len(metadata_tmp.keys())>0: # Check if something was found
        mkeys = list(metadata_tmp.keys())
        #print('-----')
        #print('Got metadata keys',metadata_tmp)
        try:
            datakey_unit = metadata_tmp['unit']
            mkeys.remove('unit')
            mkeys.insert(0,'unit')
        except:
            datakey_unit = None

        for metakey in mkeys:
            # Get the rows and columns
            try:
                rowind = all_worksheets['metadata_indices']['rows'].index(metakey)
                rowind_write = rowind + row_firstdata
            except:
                all_worksheets['metadata_indices']['rows'].append(metakey)
                rowind = all_worksheets['metadata_indices']['rows'].index(metakey)
                rowind_write = rowind + row_firstdata
                colind_datakey = 0
                # Write the metakey in the correct row
                #all_worksheets['metadata'].write(row_firstdata, 0, 'Metadata key')
                all_worksheets['metadata'].write(rowind_write, colind_datakey, metakey, all_worksheets['header_format'])

            # And finally write the metadata
            try:
                colind = all_worksheets['metadata_indices']['columns'].index(raddress_tmp_str)
                colind_write = colind + 1
            except:
                try:
                    all_worksheets['metadata_indices']['columns'].append(raddress_tmp_str)
                    colind = all_worksheets['metadata_indices']['columns'].index(raddress_tmp_str)
                    colind_write = colind + 1
                    rowind_write_col = 0
                    #all_worksheets['metadata'].write(rowind_write_col, colind_write, raddress_tmp_str,all_worksheets['header_format'])
                    #all_worksheets['metadata'].write(rowind_write_col+1, colind_write, raddress_tmp_str_full,all_worksheets['header_format'])
                    for header_row in header_rows:
                        all_worksheets['metadata'].write(header_row[0], colind_write, header_row[1],
                                                         all_worksheets['header_format'])

                except:
                    logger.info('Error',exc_info=True)


            #print('Writing metadata for address')
            datawrite = metadata_tmp[metakey]
            all_worksheets['metadata'].write(rowind_write, colind_write, datawrite)

        all_worksheets['metadata'].autofit()
        return datakey_unit

def start(device_info, config, dataqueue=None, datainqueue=None, statusqueue=None):
    funcname = __name__ + '.start()'
    logger.debug(funcname + ':Opening writing:')
    #print('Config',config)
    data_write_to_file = [] # List of columns to be written to file
    count = 0
    all_worksheets = {}
    deviceinfo_all = None
    if True:
        try:
            dtneworig = config['dt_newfile']
            dtunit = config['dt_newfile_unit']
            if(dtunit.lower() == 'seconds'):
                dtfac = 1.0
            elif(dtunit.lower() == 'hours'):
                dtfac = 3600.0
            elif(dtunit.lower() == 'days'):
                dtfac = 86400.0
            else:
                dtfac = 0
                
            dtnews = dtneworig * dtfac
            logger.info(funcname + ' Will create new file every {:d} {:s}.'.format(config['dt_newfile'],config['dt_newfile_unit']))
        except:
            logger.warning(funcname + 'Could not start',exc_info=True)
            dtnews = 0
            
        try:
            sizeneworig = config['size_newfile']
            sizeunit = config['size_newfile_unit']
            if(sizeunit.lower() == 'kb'):
                sizefac = 1000.0
            elif(sizeunit.lower() == 'mb'):            
                sizefac = 1e6
            elif(sizeunit.lower() == 'bytes'):            
                sizefac = 1 
            else:
                sizefac = 0
                
            sizenewb = sizeneworig * sizefac # Size in bytes
            logger.info(funcname + ' Will create new file every {:d} {:s}.'.format(config['size_newfile'],config['size_newfile_unit']))
        except Exception as e:
            logger.exception(e)
            sizenewb = 0  # Size in bytes
            
    try:
        config['dt_sync']
    except:
        config['dt_sync'] = 5

    bytes_written = 0
    packets_written = 0
    bytes_written_total = 0
    packets_written_total = 0
    device_worksheets = {} # The extended info of the worksheets for the devices
    device_worksheets_reduced = {} # The worksheets for the devices
    device_worksheets_indices = {}
    numworksheet = 0
    [workbook,filename,formats] = create_logfile(config,count,all_worksheets)
    worksheet_summary = all_worksheets['summary']

    data_stat = {'_deviceinfo': {}}
    data_stat['_deviceinfo']['filename'] = filename
    data_stat['_deviceinfo']['filename_full'] = os.path.realpath(filename)
    data_stat['_deviceinfo']['created'] = time.time()
    data_stat['_deviceinfo']['bytes_written'] = bytes_written
    data_stat['_deviceinfo']['packets_written'] = packets_written
    dataqueue.put(data_stat)
    count += 1

    tfile = time.time() # Save the time the file was created
    tflush = time.time() # Save the time the file was created
    tupdate = time.time() # Save the time for the update timing
    FLAG_RUN = True
    file_status = {}
    file_status_reduced = {}
    while FLAG_RUN:
        tcheck = time.time()
        # Flush file on regular basis
        if ((time.time() - tflush) > config['dt_sync']):
            #print('Flushing, not implemented (yet)')
            bytes_written = pympler.asizeof.asizeof(workbook)
            #print('Bytes written',bytes_written)
            tflush = time.time()

        time.sleep(0.05)
        while(datainqueue.empty() == False):
            try:
                data = datainqueue.get(block=False)
                packet_address = redvypr.RedvyprAddress(data)
                #print('xlsxlogger: Got data',data)

                if (data is not None):
                    [command,comdata] = data_packets.check_for_command(data, thread_uuid=device_info['thread_uuid'], add_data=True)
                    if (command is not None):
                        logger.debug('Command: {:s}'.format(str(command)))
                        if(command == 'stop'):
                            logger.debug('Stop command')
                            FLAG_RUN = False
                            break

                        if (command == 'info'):
                            logger.debug('Metadata command')
                            if packet_address.packetid == 'metadata':
                                deviceinfo_all = data['deviceinfo_all']

                #statistics = data_packets.do_data_statistics(data,statistics)
                address_format = '/h/p/d/'
                packet_address_str = packet_address.get_str(address_format)
                colindex_time = 1
                colindex_numpacket = 0
                coloffset = colindex_time + 1
                # Ignore some packages
                if data in redvypr_address.RedvyprAddress(redvypr.metadata_address):
                    logger.debug('Ignoring metadata packet')
                    continue

                try:
                    device_worksheets[packet_address_str]
                except:
                    numworksheet += 1
                    numworksheet_offset = 2
                    devicename = packet_address.devicename.replace('/','_').replace('[','_').replace(']','_').replace('\\','_').replace('*','_').replace(':','_')
                    packet_address_str_xlsx = '{:02d}_{}'.format(numworksheet,devicename)
                    if len(packet_address_str_xlsx) > 31:
                        packet_address_str_xlsx = packet_address_str_xlsx[0:31]

                    worksheet_summary.write(numworksheet_offset, 0, 'device')
                    worksheet_summary.write(numworksheet_offset, 1, 'worksheet')
                    worksheet_summary.write(numworksheet + numworksheet_offset, 0, packet_address_str)
                    worksheet_summary.write(numworksheet + numworksheet_offset, 1, packet_address_str_xlsx)
                    logger.debug('Will create workbook for {}'.format(packet_address_str_xlsx))
                    device_worksheets[packet_address_str] = workbook.add_worksheet(packet_address_str_xlsx)
                    device_worksheets_indices[packet_address_str] = {'datakeys':[],'numline':0,'colindex':{}}
                    device_worksheets_indices[packet_address_str]['worksheet'] = packet_address_str_xlsx
                    device_worksheets_reduced[packet_address_str] = "worksheet: {}, #written: {}".format(packet_address_str_xlsx,0)
                    #
                    device_worksheets[packet_address_str].write(row_address,0,'Address',all_worksheets['header_format'])
                    device_worksheets[packet_address_str].write(row_address,1,packet_address_str,all_worksheets['header_format'])
                    #
                    device_worksheets[packet_address_str].write(row_datakey, colindex_time, 'Excel time',all_worksheets['header_format'])
                    for row in range(row_host,row_datakey):
                        device_worksheets[packet_address_str].write(row, colindex_time, '',
                                                                    all_worksheets['header_format'])
                    device_worksheets[packet_address_str].write(row_dataunit, colindex_time, time_format,all_worksheets['header_format'])
                    device_worksheets[packet_address_str].write(row_datakey, colindex_numpacket, 'Numpacket',all_worksheets['header_format'])
                    device_worksheets[packet_address_str].write(row_dataunit, colindex_numpacket, '#',all_worksheets['header_format'])
                    try:
                        file_status[filename]['worksheets']
                    except:
                        file_status[filename] = {'worksheets':{}}
                    try:
                        file_status_reduced[filename]['worksheets']
                    except:
                        file_status_reduced[filename] = {'worksheets': {}}

                    data_stat = {'_deviceinfo': {}}
                    file_status[filename]['worksheets'][packet_address_str] = device_worksheets_indices[packet_address_str]
                    file_status_reduced[filename]['worksheets'] = device_worksheets_reduced
                    data_stat['_deviceinfo']['file_status'] = file_status
                    data_stat['_deviceinfo']['file_status_reduced'] = file_status_reduced
                    dataqueue.put(data_stat)

                # Write data
                if True:
                    try:
                        data['t']
                    except:
                        data['t'] = data['_redvypr']['t']

                    try:
                        numpacket = data['_redvypr']['numpacket']
                    except:
                        numpacket = -1

                    datapacket = data_packets.Datapacket(data)
                    datakeys = datapacket.datakeys(expand=config['datakey_expansionlevel'],return_type='list')
                    #print('datakeys',datakeys)
                    lineindex = row_firstdata + device_worksheets_indices[packet_address_str]['numline']
                    # Write numpacket
                    device_worksheets[packet_address_str].write(lineindex, colindex_numpacket, numpacket)
                    # Write time
                    datakeys.remove('t')
                    datakeys.insert(0,'t')
                    datatime = datetime.datetime.fromtimestamp(data['t'])
                    device_worksheets[packet_address_str].write_datetime(lineindex, colindex_time, datatime,formats['date'])
                    # Write data in datakeys
                    for datakey in datakeys:
                        raddress_datakey = redvypr_address.RedvyprAddress(data, datakey=datakey)
                        # Get metadata
                        if deviceinfo_all is not None:
                            datakey_unit = write_metadata(workbook, all_worksheets, datakey, data, deviceinfo_all, device_info, config)
                        else:
                            #print('No metadata to check for')
                            datakey_unit = None

                        if (datakey == 't') and (datakey_unit is None):
                            datakey_unit = 'unix time'

                        try:
                            colindex = device_worksheets_indices[packet_address_str]['colindex'][datakey]
                        except:
                            device_worksheets_indices[packet_address_str]['datakeys'].append(datakey)
                            colindex = len(device_worksheets_indices[packet_address_str]['datakeys']) - 1 + coloffset
                            device_worksheets_indices[packet_address_str]['colindex'][datakey] = colindex
                            header_rows = [(row_host,raddress_datakey.hostname,'hostname')]
                            header_rows.append((row_uuid, raddress_datakey.uuid,'uuid'))
                            header_rows.append((row_device, raddress_datakey.devicename,'devicename'))
                            header_rows.append((row_publisher, raddress_datakey.publisher,'publisher'))
                            header_rows.append((row_packetid, raddress_datakey.packetid,'packetid'))
                            header_rows.append((row_datakey, datakey,'datakey'))
                            for header_row in header_rows:
                                device_worksheets[packet_address_str].write(header_row[0],colindex,header_row[1],
                                                                            all_worksheets['header_format'])
                                device_worksheets[packet_address_str].write(header_row[0], 0, header_row[2],
                                                                            all_worksheets['header_format'])


                            #device_worksheets[packet_address_str].write(row_datakey,colindex,datakey,all_worksheets['header_format'])


                        if datakey_unit is not None:
                            device_worksheets[packet_address_str].write(row_dataunit, colindex, datakey_unit,all_worksheets['header_format'])
                        else:
                            datakey_unit_tmp = ''
                            device_worksheets[packet_address_str].write(row_dataunit, colindex, datakey_unit_tmp,
                                                                        all_worksheets['header_format'])
                        #print('Will write data from {} to column {} and line {}'.format(packet_address_str, colindex,lineindex))
                        datawrite = datapacket[datakey]
                        if not(isinstance(datawrite,str)) and not(isinstance(datawrite,int)) and not(isinstance(datawrite,float)):
                            #print('Datatype {} not supported, converting data to str'.format(str(type(datawrite))))
                            datawrite = str(datawrite)

                        device_worksheets[packet_address_str].write(lineindex, colindex, datawrite)

                    device_worksheets_indices[packet_address_str]['numline'] += 1
                    numline = device_worksheets_indices[packet_address_str]['numline']
                    numkeys = len(device_worksheets_indices[packet_address_str]['datakeys'])
                    worksheet_tmp = device_worksheets_indices[packet_address_str]['worksheet']
                    device_worksheets_reduced[packet_address_str] = "worksheet: {}, numkeys: {}, numlines: {}".format(
                        worksheet_tmp, numkeys, numline)
                    #size = sys.getsizeof(workbook)
                    packets_written += 1
                    #print('Size:',size)

                # Send statistics
                if ((time.time() - tupdate) > config['dt_update']):
                    tupdate = time.time()
                    data_stat = {'_deviceinfo': {}}
                    data_stat['_deviceinfo']['filename'] = filename
                    data_stat['_deviceinfo']['filename_full'] = os.path.realpath(filename)
                    data_stat['_deviceinfo']['bytes_written'] = bytes_written
                    data_stat['_deviceinfo']['packets_written'] = packets_written
                    file_status[filename]['worksheets'] = device_worksheets_indices
                    data_stat['_deviceinfo']['file_status'] = file_status
                    data_stat['_deviceinfo']['file_status_reduced'] = file_status_reduced
                    dataqueue.put(data_stat)

            except:
                logger.debug('Could not write data',exc_info=True)
                #logger.exception(e)
                #logger.debug(funcname + ':Exception:' + str(e))
                # print(data)

        if True: # Check if a new file should be created, close the old one and write the header
            file_age = tcheck - tfile
            FLAG_TIME = (dtnews > 0) and (file_age >= dtnews)
            FLAG_SIZE = (sizenewb > 0) and (bytes_written >= sizenewb)
            if FLAG_TIME or FLAG_SIZE or (FLAG_RUN == False):
                # Autofit
                #print('Autofit')
                if True:
                    try:
                        worksheet_summary.autofit()
                    except:
                        logger.info(funcname + 'Could not autofit summary',exc_info=True)
                    for w in device_worksheets:
                        try:
                            device_worksheets[w].autofit()
                        except:
                            logger.info(funcname + 'Could not autofit {}'.format(w), exc_info=True)
                # Make the Excel time column wider, autofit does not work properly here
                #device_worksheets[w].set_column(colindex_time, colindex_time, 25)
                workbook.close()
                data_stat = {'_deviceinfo': {}}
                data_stat['_deviceinfo']['filename'] = filename
                data_stat['_deviceinfo']['filename_full'] = os.path.realpath(filename)
                data_stat['_deviceinfo']['closed'] = time.time()
                data_stat['_deviceinfo']['bytes_written'] = bytes_written
                data_stat['_deviceinfo']['packets_written'] = packets_written
                dataqueue.put(data_stat)
                if FLAG_RUN:
                    tfile = tcheck
                    bytes_written = 0
                    packets_written = 0
                    #bytes_written_total = 0
                    #packets_written_total = 0
                    device_worksheets = {}  # The worksheets for the devices
                    device_worksheets_reduced = {}  # The worksheets for the devices
                    device_worksheets_indices = {}
                    numworksheet = 0
                    [workbook, filename, formats] = create_logfile(config, count, all_worksheets)
                    worksheet_summary = all_worksheets['summary']
                    data_stat = {'_deviceinfo': {}}
                    data_stat['_deviceinfo']['filename'] = filename
                    data_stat['_deviceinfo']['filename_full'] = os.path.realpath(filename)
                    data_stat['_deviceinfo']['created'] = time.time()
                    data_stat['_deviceinfo']['bytes_written'] = bytes_written
                    data_stat['_deviceinfo']['packets_written'] = packets_written
                    dataqueue.put(data_stat)




class Device(RedvyprDevice):
    """
    xlsxwriter device
    """

    def __init__(self, **kwargs):
        """
        """
        funcname = __name__ + '__init__()'
        super(Device, self).__init__(**kwargs)

    def thread_start(self, *args,**kwargs):
        """
        Custom thread start that clear the queue if config wants to and send a deviceinfo into the datainqueue

        Parameters
        ----------
        args
        kwargs

        Returns
        -------

        """
        funcname = __name__ + '.thread_start():'
        self.logger.debug(funcname)
        compacket = self.redvypr.get_metadata_commandpacket()
        # Let the thread start, it will clear the queue
        super().thread_start(*args,**kwargs)
        self.datainqueue.put(compacket)


#
#
# The init widget
#
#
class initDeviceWidget(QtWidgets.QWidget):
    connect      = QtCore.pyqtSignal(RedvyprDevice) # Signal requesting a connect of the datainqueue with available dataoutqueues of other devices
    def __init__(self,device=None):
        super(QtWidgets.QWidget, self).__init__()
        layout        = QtWidgets.QGridLayout(self)
        #print('Hallo,device config',device.config)
        self.device   = device
        self.redvypr  = device.redvypr
        self.label    = QtWidgets.QLabel("xlsx logger setup")
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setStyleSheet(''' font-size: 24px; font: bold''')
        self.config_widgets= [] # A list of all widgets that can only be used of the device is not started yet
        # Input output widget
        self.inlabel  = QtWidgets.QLabel("Input")
        self.inlist   = QtWidgets.QListWidget()
        #
        self.adddeviceinbtn   = QtWidgets.QPushButton("Subscribe")
        self.adddeviceinbtn.clicked.connect(self.con_clicked)
        self.adddeviceinbtn.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        # The output widgets
        self.outlabel        = QtWidgets.QLabel("Logfile")
        self.outfilename     = QtWidgets.QLineEdit()
        # Checkboxes
        self.prefix_check    = QtWidgets.QCheckBox('Prefix')
        self.date_check      = QtWidgets.QCheckBox('Date/Time')
        self.count_check     = QtWidgets.QCheckBox('Counter')
        self.postfix_check   = QtWidgets.QCheckBox('Postfix')
        self.extension_check = QtWidgets.QCheckBox('Extension')

        try:
            filename = self.device.custom_config['filename']
        except:
            filename = ''

        self.outfilename.setText(filename)
        
        self.folderbtn   = QtWidgets.QPushButton("Folder")
        self.config_widgets.append(self.folderbtn)
        self.folderbtn.clicked.connect(self.get_datafolder)
        
        # The rest
        #self.conbtn = QtWidgets.QPushButton("Connect logger to devices")
        #self.conbtn.clicked.connect(self.con_clicked)        
        self.startbtn = QtWidgets.QPushButton("Start logging")
        self.startbtn.clicked.connect(self.start_clicked)
        self.startbtn.setCheckable(True)
        self.startbtn.setSizePolicy(QtWidgets.QSizePolicy.Preferred,QtWidgets.QSizePolicy.Expanding)


        # Delta t for new file
        edit = QtWidgets.QLineEdit(self)
        onlyInt = QtGui.QIntValidator()
        edit.setValidator(onlyInt)
        self.dt_newfile = edit
        self.dt_newfile.setToolTip('Create a new file every N seconds.\nFilename is "filenamebase"_yyyymmdd_HHMMSS_count."ext".\nUse 0 to disable feature.')
        try:
            self.dt_newfile.setText(str(self.device.custom_config['dt_newfile']))
        except Exception as e:
            self.dt_newfile.setText('0')
            
        # Delta t for new file
        edit = QtWidgets.QLineEdit(self)
        onlyInt = QtGui.QIntValidator()
        edit.setValidator(onlyInt)
        self.size_newfile = edit
        self.size_newfile.setToolTip('Create a new file if N bytes of RAM are used.\nFilename is "filenamebase"_yyyymmdd_HHMMSS_count."ext".\nUse 0 to disable feature.')
        try:
            self.size_newfile.setText(str(self.device.custom_config.size_newfile))
        except Exception as e:
            self.size_newfile.setText('0')
            
        self.newfiletimecombo = QtWidgets.QComboBox()
        times = typing.get_args(self.device.custom_config.model_fields['dt_newfile_unit'].annotation)
        timeunit = self.device.custom_config.dt_newfile_unit
        for t in times:
            self.newfiletimecombo.addItem(t)

        index = self.newfiletimecombo.findText(timeunit)
        self.newfiletimecombo.setCurrentIndex(index)
        #self.newfiletimecombo.addItem('None')
        #self.newfiletimecombo.addItem('seconds')
        #self.newfiletimecombo.addItem('hours')
        #self.newfiletimecombo.addItem('days')
        #self.newfiletimecombo.setCurrentIndex(1)
            
        self.newfilesizecombo = QtWidgets.QComboBox()
        self.newfilesizecombo.addItem('None')
        self.newfilesizecombo.addItem('bytes')
        self.newfilesizecombo.addItem('kB')
        self.newfilesizecombo.addItem('MB')
        self.newfilesizecombo.setCurrentIndex(3)
        
        sizelabel = QtWidgets.QLabel('New file after')
         # File change layout
        self.newfilewidget = QtWidgets.QWidget()
        self.newfilelayout = QtWidgets.QFormLayout(self.newfilewidget)
        self.newfilelayout.addRow(sizelabel)
        self.newfilelayout.addRow(self.dt_newfile,self.newfiletimecombo)
        self.newfilelayout.addRow(self.size_newfile,self.newfilesizecombo)
        
        # Filenamelayout
        self.folder_text = QtWidgets.QLineEdit('')
        self.extension_text = QtWidgets.QLineEdit('redvypr_raw')
        self.prefix_text = QtWidgets.QLineEdit('')
        self.date_text = QtWidgets.QLineEdit('%Y-%m-%d_%H%M%S')
        self.count_text = QtWidgets.QLineEdit('04d')
        self.postfix_text = QtWidgets.QLineEdit('')

        self.prefix_check = QtWidgets.QCheckBox('Prefix')
        self.date_check = QtWidgets.QCheckBox('Date/Time')
        self.count_check = QtWidgets.QCheckBox('Counter')
        self.postfix_check = QtWidgets.QCheckBox('Postfix')
        self.extension_check = QtWidgets.QCheckBox('Extension')
        # The outwidget
        self.outwidget = QtWidgets.QWidget()
        self.outlayout = QtWidgets.QGridLayout(self.outwidget)
        # Datafolder lineedit
        self.outlayout.addWidget(self.folderbtn, 0, 0)
        self.outlayout.addWidget(self.folder_text, 0, 1,1,4)
        # Checkboxes
        self.outlayout.addWidget(self.prefix_check, 1, 0)
        self.outlayout.addWidget(self.date_check, 1, 1)
        self.outlayout.addWidget(self.count_check, 1, 2)
        self.outlayout.addWidget(self.postfix_check, 1, 3)
        self.outlayout.addWidget(self.extension_check, 1, 4)

        self.outlayout.addWidget(self.prefix_text, 2, 0)
        self.outlayout.addWidget(self.date_text, 2, 1)
        self.outlayout.addWidget(self.count_text, 2, 2)
        self.outlayout.addWidget(self.postfix_text, 2, 3)
        self.outlayout.addWidget(self.extension_text, 2, 4)

        self.outlayout.addWidget(self.newfilewidget,4,0,1,4)

        #self.outlayout.addStretch(1)
            
        layout.addWidget(self.label,0,0,1,2)
        #layout.addWidget(self.inlabel,1,0)
        #layout.addWidget(self.inlist,2,0)
        layout.addWidget(self.outlabel,1,0)
        layout.addWidget(self.outwidget,2,0)
        layout.addWidget(self.adddeviceinbtn, 5, 0)
        layout.addWidget(self.startbtn,6,0)

        self.config_to_widgets()
        self.connect_widget_signals()
        # Connect the signals that notify a change of the connection
        self.device.redvypr.device_status_changed_signal.connect(self.update_device_list)
        #self.redvypr.devices_connected.connect
        self.device.subscription_changed_signal.connect(self.update_device_list)

        self.statustimer = QtCore.QTimer()
        self.statustimer.timeout.connect(self.update_buttons)
        self.statustimer.start(500)


    def connect_widget_signals(self,connect=True):
        """
        Connects the signals of the widgets such that an update of the config is done

        Args:
            connect:

        Returns:

        """
        funcname = self.__class__.__name__ + '.connect_widget_signals():'
        logger.debug(funcname)
        if(connect):
            self.prefix_check.stateChanged.connect(self.update_device_config)
            self.postfix_check.stateChanged.connect(self.update_device_config)
            self.date_check.stateChanged.connect(self.update_device_config)
            self.count_check.stateChanged.connect(self.update_device_config)
            self.extension_check.stateChanged.connect(self.update_device_config)
            self.prefix_text.editingFinished.connect(self.update_device_config)
            self.postfix_text.editingFinished.connect(self.update_device_config)
            self.date_text.editingFinished.connect(self.update_device_config)
            self.count_text.editingFinished.connect(self.update_device_config)
            self.extension_text.editingFinished.connect(self.update_device_config)
            self.newfilesizecombo.currentIndexChanged.connect(self.update_device_config)
            self.newfiletimecombo.currentIndexChanged.connect(self.update_device_config)
        else:
            self.prefix_check.stateChanged.disconnect()
            self.postfix_check.stateChanged.disconnect()
            self.date_check.stateChanged.disconnect()
            self.count_check.stateChanged.disconnect()
            self.extension_check.stateChanged.disconnect()
            self.prefix_text.editingFinished.disconnect()
            self.postfix_text.editingFinished.disconnect()
            self.date_text.editingFinished.disconnect()
            self.count_text.editingFinished.disconnect()
            self.extension_text.editingFinished.disconnect()
            self.newfilesizecombo.currentIndexChanged.disconnect()
            self.newfiletimecombo.currentIndexChanged.disconnect()


    def get_datafolder(self):
        funcname = self.__class__.__name__ + '.get_datafolder():'
        logger.debug(funcname)
        retdata = QtWidgets.QFileDialog.getExistingDirectory(self,"Choose datafolder")
        #print('Datafolder',retdata)
        datafolder = retdata
        if datafolder:
            self.folder_text.setText(datafolder)
            
    def con_clicked(self):
        funcname = self.__class__.__name__ + '.con_clicked():'
        logger.debug(funcname)
        button = self.sender()
        if(button == self.adddeviceinbtn):
            self.connect.emit(self.device) # The connect signal is connected with connect_device that will open a subscribe/connect widget
            self.update_device_list()

    def update_datastream_table(self):
        """

        """
        # columns = ['Columnnr.', 'Datakey', 'Device', 'Format', 'Full Address String']
        funcname = self.__class__.__name__ + '.update_datastream_table():'
        logger.debug(funcname)
        datastreams_subscribed = self.device.get_subscribed_datastreams()
        logger.debug('datastreams subscribed {}'.format(datastreams_subscribed))
        self.datastreamtable.clear()
        self.datastreamtable.setRowCount(len(datastreams_subscribed))
        for i,d in enumerate(datastreams_subscribed):
            dadr = redvypr_address(d)
            item = QtWidgets.QTableWidgetItem(d)
            self.datastreamtable.setItem(i,4,item)

        self.datastreamtable.resizeColumnsToContents()

    def update_device_list(self):
        funcname = self.__class__.__name__ + '.update_device_list():'
        logger.debug(funcname)
        #print('Devices',devicestr_provider,devicestr_receiver)
        raddresses = self.device.get_subscribed_deviceaddresses()
        #print('Deviceaddresses',raddresses)
        self.inlist.clear()
        for raddr in raddresses:
            self.inlist.addItem(raddr.address_str)


    def config_to_widgets(self):
        """
        Updates the widgets according to the device config

        Returns:

        """
        funcname = self.__class__.__name__ + '.config_to_widgets():'
        logger.debug(funcname)

        config = self.device.custom_config
        #print('config',config)
        self.dt_newfile.setText(str(config.dt_newfile))
        for i in range(self.newfiletimecombo.count()):
            self.newfiletimecombo.setCurrentIndex(i)
            if(self.newfiletimecombo.currentText().lower() == config.dt_newfile_unit):
                break

        for i in range(self.newfilesizecombo.count()):
            self.newfilesizecombo.setCurrentIndex(i)
            if (self.newfilesizecombo.currentText().lower() == config.size_newfile_unit):
                break

        self.size_newfile.setText(str(config.size_newfile))

        if len(config.datafolder)>0:
            self.folder_text.setText(config.datafolder)
        # Update filename and checkboxes
        filename_all = []
        filename_all.append([config.fileextension,self.extension_text,self.extension_check])
        filename_all.append([config.fileprefix,self.prefix_text,self.prefix_check])
        filename_all.append([config.filepostfix,self.postfix_text,self.postfix_check])
        filename_all.append([config.filedateformat,self.date_text,self.date_check])
        filename_all.append([config.filecountformat,self.count_text,self.count_check])
        for i in range(len(filename_all)):
            widgets = filename_all[i]
            if(len(widgets[0])==0):
                widgets[2].setChecked(False)
                widgets[1].setText('')
            else:
                widgets[2].setChecked(True)
                widgets[1].setText(widgets[0])

    def widgets_to_config(self,config):
        """
        Reads the widgets and creates a config
        Returns:
            config: Config dictionary
        """
        funcname = self.__class__.__name__ + '.widgets_to_config():'
        logger.debug(funcname)
        config.dt_newfile = int(self.dt_newfile.text())
        config.dt_newfile_unit = self.newfiletimecombo.currentText()
        config.size_newfile = int(self.size_newfile.text())
        config.size_newfile_unit = self.newfilesizecombo.currentText()

        config.datafolder = self.folder_text.text()

        if(self.extension_check.isChecked()):
            config.fileextension = self.extension_text.text()
        else:
            config.fileextension = ''

        if(self.prefix_check.isChecked()):
            config.fileprefix = self.prefix_text.text()
        else:
            config.fileprefix = ''

        if(self.postfix_check.isChecked()):
            config.filepostfix = self.postfix_text.text()
        else:
            config.filepostfix = ''

        if(self.date_check.isChecked()):
            config.filedateformat = self.date_text.text()
        else:
            config.filedateformat = ''

        if(self.count_check.isChecked()):
            config.filecountformat = self.count_text.text()
        else:
            config.filecountformat = ''

        #print('Config',config)
        return config

    def update_device_config(self):
        """
        Updates the device config based on the widgets
        Returns:

        """
        funcname = self.__class__.__name__ + '.update_device_config():'
        logger.debug(funcname)
        self.widgets_to_config(self.device.custom_config)

    def start_clicked(self):
        funcname = self.__class__.__name__ + '.start_clicked():'
        logger.debug(funcname)
        button = self.sender()
        if button.isChecked():
            logger.debug(funcname + "button pressed")
            self.update_device_config()
            self.device.thread_start()
        else:
            logger.debug(funcname + 'button released')
            self.device.thread_stop()

            
    def update_buttons(self):
            """ Updating all buttons depending on the thread status (if its alive, graying out things)
            """

            status = self.device.get_thread_status()
            thread_status = status['thread_running']

            # Running
            if(thread_status):
                self.startbtn.setText('Stop')
                self.startbtn.setChecked(True)
                for w in self.config_widgets:
                    w.setEnabled(False)
            # Not running
            else:
                self.startbtn.setText('Start')
                for w in self.config_widgets:
                    w.setEnabled(True)
                    
                # Check if an error occured and the startbutton 
                if(self.startbtn.isChecked()):
                    self.startbtn.setChecked(False)
                #self.conbtn.setEnabled(True)


class displayDeviceWidget(QtWidgets.QWidget):
    def __init__(self,device=None):
        super(QtWidgets.QWidget, self).__init__()
        self.device = device
        layout          = QtWidgets.QVBoxLayout(self)
        hlayout         = QtWidgets.QHBoxLayout()
        self.filetable  = QtWidgets.QTableWidget()
        headers = ['Filename','Bytes written','Packets written','Date created','Date closed','Full filepath']
        self.filetable.setColumnCount(len(headers))
        self.filetable.setHorizontalHeaderLabels(headers)
        self.filetable.resizeColumnsToContents()
        self.filelab = QtWidgets.QLabel("File: ")
        self.byteslab = QtWidgets.QLabel("Bytes written: ")
        self.packetslab = QtWidgets.QLabel("Packets written: ")
        # Table that displays all datastreams and the format as it is written to the file
        self.deviceinfoQtree = redvypr.widgets.pydanticConfigWidget.dictQTreeWidget(dataname='file status', show_datatype=False)
        # Update layout
        updatelayout = QtWidgets.QHBoxLayout()
        self.update_auto = QtWidgets.QCheckBox('Autoupdate file status')
        self.update_auto.setChecked(True)
        updatelayout.addWidget(self.update_auto)
        self.update_show_all_files = QtWidgets.QRadioButton('Show all files')
        #self.update_show_all_files.setChecked(True)
        self.update_show_cur_file = QtWidgets.QRadioButton('Show current file')
        self.update_show_cur_file.setChecked(True)
        show_group=QtWidgets.QButtonGroup()
        show_group.addButton(self.update_show_all_files)
        show_group.addButton(self.update_show_cur_file)
        self.show_group = show_group
        updatelayout.addWidget(self.update_show_all_files)
        updatelayout.addWidget(self.update_show_cur_file)

        self.update_show_full = QtWidgets.QRadioButton('Show all information')
        #self.update_show_full.setChecked(True)
        self.update_show_red = QtWidgets.QRadioButton('Show reduced information')
        self.update_show_red.setChecked(True)
        showred_group = QtWidgets.QButtonGroup()
        showred_group.addButton(self.update_show_full)
        showred_group.addButton(self.update_show_red)
        self.showred_group = showred_group
        updatelayout.addWidget(self.update_show_full)
        updatelayout.addWidget(self.update_show_red)



        hlayout.addWidget(self.byteslab)
        hlayout.addWidget(self.packetslab)
        layout.addWidget(self.filelab)        
        layout.addLayout(hlayout)
        layout.addWidget(self.filetable)
        layout.addWidget(self.deviceinfoQtree)
        # Add update options
        layout.addLayout(updatelayout)
        #self.text.insertPlainText("hallo!")

    def update_qtreewidget(self, data):
        funcname = __name__ + '.update_qtreewidget()'
        logger.debug(funcname)
        if self.update_auto.isChecked():
            devinfo = data
            #devinfo = self.device.get_device_info(address=self.device.address_str)
            # devinfo = self.device.get_device_info()
            # print('Deviceinfo!!!!', devinfo)
            # print('__________')

            filename = devinfo['_deviceinfo']['filename']

            if self.update_show_red.isChecked():
                #print('Hallo show reduced is checked')
                file_status = devinfo['_deviceinfo']['file_status_reduced']
            else:
                #print('Hallo show full is checked')
                file_status = devinfo['_deviceinfo']['file_status']

            if self.update_show_cur_file.isChecked():
                #print('Hallo current file is checked')
                file_status = file_status[filename]
            else:
                #print('Hallo is not checked')
                pass


            # Update the qtree
            # https://stackoverflow.com/questions/9364754/remembering-scroll-value-of-a-qtreewidget-in-pyqt?rq=3
            bar = self.deviceinfoQtree.verticalScrollBar()
            yScroll = bar.value()
            self.deviceinfoQtree.reload_data(file_status)
            self.deviceinfoQtree.verticalScrollBar().setSliderPosition(yScroll)

    def update_data(self,data):
        funcname = __name__ + '.update()'
        try:
            data['_deviceinfo']
        except:
            return

        # Update qtree
        try:
            # Test if the file status has changed, if yes make an update
            file_status_tmp = data['_deviceinfo']['file_status']
            try:
                self.update_qtreewidget(data)
            except:
                logger.debug(funcname+ 'Qtreeupdate',exc_info=True)

        except Exception as e:
            #logger.debug(funcname, exc_info=True)
            pass

        try:
            #print('data',data)
            try:
                filename_table = self.filetable.item(0,0).text()
            except:
                filename_table = ''

            try:
                data['_deviceinfo']['filename']
                FLAG_FILEUPDATE = True
            except:
                FLAG_FILEUPDATE = False

            try:
                data['_deviceinfo']['csvcolumns']
                FLAG_CSVCOLUMNUPDATE = True
            except:
                FLAG_CSVCOLUMNUPDATE = False

            if FLAG_CSVCOLUMNUPDATE:
                #print('csv column update')
                csvcolumns = data['_deviceinfo']['csvcolumns']
                nrows = len(csvcolumns)
                self.datastreamtable.setRowCount(nrows)
                iaddr = self.datastreamtable_columns.index('Address')
                idatatype = self.datastreamtable_columns.index('Datatype')
                idatakey = self.datastreamtable_columns.index('Datakey')
                idevice = self.datastreamtable_columns.index('Device')
                iformat = self.datastreamtable_columns.index('Format')
                icolnr = self.datastreamtable_columns.index('Columnnr.')
                #self.datastreamtable_columns = ['Columnnr.', 'Datakey', 'Device', 'Datatype', 'Format', 'Address']

                for i,c in enumerate(csvcolumns):
                    item = QtWidgets.QTableWidgetItem(str(i))
                    self.datastreamtable.setItem(i, icolnr, item)

                    datatype = data['_deviceinfo']['header_datatype'][i]
                    item = QtWidgets.QTableWidgetItem(datatype)
                    self.datastreamtable.setItem(i, idatatype, item)

                    datakey = data['_deviceinfo']['header_datakeys'][i]
                    item = QtWidgets.QTableWidgetItem(datakey)
                    self.datastreamtable.setItem(i, idatakey, item)

                    dataformat = data['_deviceinfo']['csvformat'][i]
                    item = QtWidgets.QTableWidgetItem(dataformat)
                    self.datastreamtable.setItem(i, iformat, item)

                    datadevice = data['_deviceinfo']['header_device'][i]
                    item = QtWidgets.QTableWidgetItem(datadevice)
                    self.datastreamtable.setItem(i, idevice, item)

                    item = QtWidgets.QTableWidgetItem(str(c))
                    self.datastreamtable.setItem(i, iaddr, item)

                self.datastreamtable.resizeColumnsToContents()
            if FLAG_FILEUPDATE:
                #print('Filename table',filename_table,data['_deviceinfo']['filename'])
                try:
                    tclose = datetime.datetime.fromtimestamp(data['_deviceinfo']['closed']).strftime(
                        '%d-%m-%Y %H:%M:%S')
                    item = QtWidgets.QTableWidgetItem(tclose)
                    self.filetable.setItem(0, 4, item)
                except:
                    pass

                if filename_table != data['_deviceinfo']['filename']:
                    self.filetable.insertRow(0)
                    try:
                        #headers = ['Filename', 'Date created', 'Bytes written', 'Date closed']
                        item = QtWidgets.QTableWidgetItem(data['_deviceinfo']['filename'])
                        self.filetable.setItem(0,0,item)
                        item = QtWidgets.QTableWidgetItem(data['_deviceinfo']['filename_full'])
                        self.filetable.setItem(0, 5, item)
                        tcreate = datetime.datetime.fromtimestamp(data['_deviceinfo']['created']).strftime('%d-%m-%Y %H:%M:%S')
                        item = QtWidgets.QTableWidgetItem(tcreate)
                        self.filetable.setItem(0, 3, item)
                    except Exception as e:
                        logger.exception(e)

                item = QtWidgets.QTableWidgetItem(str(data['_deviceinfo']['bytes_written']))
                self.filetable.setItem(0, 1, item)
                item = QtWidgets.QTableWidgetItem(str(data['_deviceinfo']['packets_written']))
                self.filetable.setItem(0, 2, item)
                self.filetable.resizeColumnsToContents()
                self.filelab.setText("File: {:s}".format(data['_deviceinfo']['filename']))
                self.byteslab.setText("Bytes written: {:d}".format(data['_deviceinfo']['bytes_written']))
                self.packetslab.setText("Packets written: {:d}".format(data['_deviceinfo']['packets_written']))
        except Exception as e:
            logger.exception(e)


        #self.text.insertPlainText(str(data['_deviceinfo']['data']))
        



import logging
import numpy as np
import sys
from redvypr.redvypr_address import RedvyprAddress
import redvypr.data_packets as data_packets


logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('average_data')
logger.setLevel(logging.DEBUG)

class average_data():
    avg_data = {}
    def __init__(self, data, datakeys_avg, loggerconfig):
        self.loggerconfig = loggerconfig
        sn = data['sn']
        self.sn = sn
        self.macstr = sn
        data_addr = RedvyprAddress(data)
        devname = data_addr.devicename
        devsplit = devname.rsplit('_',maxsplit=1)[0]
        self.devicename = devsplit + '_AVG_' + self.macstr
        self.datakeys_avg = datakeys_avg
        self.avg_data = {}
        self.avg_tstart = {}
    def average_data(self,data):
        """
        Abveraging data
        """
        if self.loggerconfig is not None:
            datapacket_avg = None
            # Loop over all average intervals
            for avg_s in self.loggerconfig.avg_data:
                print('Checking for average {}'.format(avg_s))
                try:
                    self.avg_data[avg_s]
                    self.avg_tstart[avg_s]
                except:
                    self.avg_data[avg_s] = []
                    self.avg_tstart[avg_s] = data['t']

                dt = data['t'] - self.avg_tstart[avg_s]
                # Average reached
                if dt >= avg_s:
                    avg_str = str(avg_s) + 's'
                    if len(self.avg_data[avg_s]) > 0:
                        # devicename = datapacket_HFSI['devicename'] + '_avg'


                        datapacket_avg = data_packets.create_datadict(device=self.devicename, tu=data['_redvypr']['t'], hostinfo=self.device_info['hostinfo'])

                        datapacket_avg['type'] = self.data['type']
                        datapacket_avg['sn'] = self.macstr
                        print('Average reached')
                        for parameter in self.datakeys_avg:
                            avgdata = []
                            for d in self.avg_data[avg_s]:
                                avgdata.append(d[parameter])

                            if parameter == 't':
                                datapacket_avg[parameter] = np.mean(avgdata)
                            else:
                                datapacket_avg[parameter + '_avg_{}'.format(avg_str)] = np.mean(avgdata)
                                datapacket_avg[parameter + '_std_{}'.format(avg_str)] = np.std(avgdata)
                                datapacket_avg[parameter + '_n_{}'.format(avg_str)] = len(avgdata)

                        print('Publishing avg:{} packet {}'.format(avg_str, str(datapacket_HFSI_avg)))
                    else:
                        logger.debug('No data to average')

                    self.avg_data[avg_s] = [data]
                    self.avg_tstart[avg_s] = data['t']
                    return datapacket_avg
                else:
                    self.avg_data[avg_s].append(data)
                    return None


import pydantic
import typing
import numpy as np
import datetime
import dateutil
import uuid
from redvypr.redvypr_address import RedvyprAddress


def get_date_from_calibration(calibration, parameter, return_str = False, strformat = '%Y-%m-%d %H:%M:%S'):
    """
    Searches within the calibration for a calibration date. This is a helper function as the date made be at different locations
    """
    funcname = __name__ + '.get_date_from_calibration():'
    try:
        coeff = calibration['parameter'][parameter]
    except:
        coeff = None
    # Get the calibration data, that can be either in each of the parameters, or as a key for all parameter
    datestr = '1970-01-01 00:00:00'
    try:
        datestr = str(coeff['date'])
    except:
        try:
            datestr = str(calibration['date'])
        except:
            datestr = str(calibration.date)

    #logger.debug(funcname + ' dates {:s}'.format(datestr))
    print(funcname + ' dates {:s}'.format(datestr))
    dateformats = ['%d.%m.%Y %H:%M:%S', '%Y-%m-%d %H:%M:%S','%Y-%m-%d %H:%M:%S.%f', '%Y-%m-%d_%H-%M-%S']
    for df in dateformats:
        try:
            td = datetime.datetime.strptime(datestr, df)  # '18.07.2023 13:42:16'
        except:
            td = None

    # Last resort, dateutil
    if td is None:
        try:
            td = dateutil.parser.parse(datestr)
        except:
            td = None

    if return_str and (td is not None):
        return td.strftime(strformat)
    else:
        return td


class calibration_const(pydantic.BaseModel):
    """
    Calibration model for a NTC sensor
    """
    structure_version: str = '1.0'
    calibration_type: typing.Literal['const'] = 'const'
    address_apply: typing.Optional[RedvyprAddress] = pydantic.Field(default=None,
                                               escription='The address of the datakey in the datapacket data calibration shall be applied to')
    datakey_result: typing.Optional[str] = pydantic.Field(default=None, description='The keyname of the output of the calibration. Note that this is the basekey of the datadictionary.')
    parameter: RedvyprAddress = pydantic.Field(default = RedvyprAddress('/k:PARA_CONST'),description='The address of calibrated parameter in the datapacket', editable=False)
    sn: str = '' # The serial number of the sensor
    sensor_model: str = ''  # The sensor model
    coeff: float = pydantic.Field(default = 1.0)
    unit: str = 'NA'
    unit_input: str = 'NA'
    date: datetime.datetime = pydantic.Field(default=datetime.datetime(1970,1,1,0,0,0), description='The calibration date')
    calibration_id: str = pydantic.Field(default='', description='ID of the calibration, can be choosen by the user')
    calibration_uuid: str = pydantic.Field(default_factory=lambda: uuid.uuid4().hex,
                                         description='uuid of the calibration, can be choosen by the user')
    comment: typing.Optional[str] = None

    def raw2data(self, raw_data):
        data = raw_data * self.coeff
        return data

class calibration_poly(pydantic.BaseModel):
    """
    Calibration model for a sensor with a polynomial response
    """
    structure_version: str = '1.0'
    calibration_type: typing.Literal['poly'] = 'poly'
    parameter: str = pydantic.Field(default = 'XYZ',description='The calibrated parameter')
    sn: str = '' # The serial number of the sensor
    sensor_model: str = ''  # The sensor model
    coeff: list = pydantic.Field(default = [1.0, 0], description='The calibration polynomial. The first entry is the one with the highest exponent, the last the constant: y = coeff[-1] + x * coeff[-2] + x**2 * coeff[-3]')
    unit: str = 'NA'
    unit_input: str = 'NA'
    date: datetime.datetime = pydantic.Field(default=datetime.datetime(1970,1,1,0,0,0), description='The calibration date')
    calibration_id: str = pydantic.Field(default='', description='ID of the calibration, can be choosen by the user')
    calibration_uuid: str = pydantic.Field(default_factory=lambda: uuid.uuid4().hex,
                                         description='uuid of the calibration, can be choosen by the user')
    comment: typing.Optional[str] = None

    def raw2data(self, raw_data):
        data = np.polyval(self.coeff,raw_data)
        return data


class calibration_HF(pydantic.BaseModel):
    """
    Calibration model for a heatflow sensor
    """
    structure_version: str = '1.0'
    calibration_type: typing.Literal['heatflow'] = 'heatflow'
    parameter: str = 'HF'
    sn: str = '' # The serial number of the sensor
    sensor_model: str = ''  # The sensor model
    coeff: float = np.nan
    coeff_std: typing.Optional[float] = None
    unit:  str = 'W m-2 mV-1'
    unit_input: str = 'mV'
    date: datetime.datetime = pydantic.Field(default=datetime.datetime(1970, 1, 1, 0, 0, 0),
                                             description='The calibration date')
    calibration_id: str = pydantic.Field(default='',
                                         description='ID of the calibration, can be choosen by the user.')
    calibration_uuid: str = pydantic.Field(default_factory=lambda: uuid.uuid4().hex, description='uuid of the calibration, can be choosen by the user. ID should be unique')
    comment: typing.Optional[str] = None

class calibration_NTC(pydantic.BaseModel):
    """
    Calibration model for a NTC sensor
    """
    structure_version: str = '1.0'
    calibration_type: typing.Literal['ntc'] = 'ntc'
    address_apply: typing.Optional[RedvyprAddress] = pydantic.Field(default=None,
                                               escription='The address of the datakey in the datapacket data calibration shall be applied to')
    datakey_result: typing.Optional[str] = pydantic.Field(default='{datakey}_cal_ntc', description='The keyname of the output of the calibration. Note that this is the basekey of the datadictionary.')
    #parameter: str = pydantic.Field(default = 'NTC_NUMXYZ',description='The calibrated parameter, this links the calibration to a specific sensor, i.e. NTC0 or NTC_A[10]')
    parameter: RedvyprAddress = pydantic.Field(default=RedvyprAddress('NTC[10]'),
                                    description='The calibrated parameter, this links the calibration to a specific sensor, i.e. NTC0 or NTC_A[10]')
    sn: str = '' # The serial number of the sensor
    sensor_model: str = ''  # The sensor model
    Toff: float = 273.15 # Offset between K and degC
    coeff: list = pydantic.Field(default = [np.nan, np.nan, np.nan, np.nan])
    #coeff_std: typing.Optional[float] = None
    unit: str = 'degC'
    unit_input: str = 'Ohm'
    date: datetime.datetime = pydantic.Field(default=datetime.datetime(1970,1,1,0,0,0), description='The calibration date')
    calibration_id: str = pydantic.Field(default='', description='ID of the calibration, can be choosen by the user')
    calibration_uuid: str = pydantic.Field(default_factory=lambda: uuid.uuid4().hex,
                                         description='uuid of the calibration, can be choosen by the user')
    comment: typing.Optional[str] = None

    def raw2data(self,data):
        """
        Calculate the temperature based on the calibration and a resistance using a polynome
        """
        P_R = self.coeff
        Toff = self.Toff
        T_1 = np.polyval(P_R, np.log(data))
        T = 1 / T_1 - Toff
        return T


calibration_models = []
calibration_models.append(calibration_const)
calibration_models.append(calibration_poly)
calibration_models.append(calibration_HF)
calibration_models.append(calibration_NTC)



import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import sys
import pynmea2
import pydantic
import typing
import copy

import redvypr.data_packets
import redvypr.data_packets as data_packets
from redvypr.device import RedvyprDevice
from redvypr.data_packets import check_for_command



logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('nmeaparser')
logger.setLevel(logging.DEBUG)

description = 'Parses NMEA data strings'

redvypr_devicemodule = True
class DeviceBaseConfig(pydantic.BaseModel):
    publishes: bool = True
    subscribes: bool = False
    description: str = 'NMEA0183 parsing device'

class DeviceCustomConfig(pydantic.BaseModel):
    dt_status: float = 2.0
    messages:list = pydantic.Field(default=[])
    datakey: str = pydantic.Field(default='',description='The datakey to look for NMEA data, if empty scan all fields and use the first match')

def start(device_info, config=None, dataqueue=None, datainqueue=None, statusqueue=None):
    """ 
    """
    funcname = __name__ + '.start()'

    # Some variables for status
    tstatus = time.time()

    dtstatus = config['dt_status']

    #
    npackets = 0  # Number packets received
    while True:
        data = datainqueue.get()
        command = check_for_command(data)
        if(command is not None):
            if command == 'stop':
                logger.debug('Got a stop command, stopping thread')
                break

        #print('Data',data)
        # Check if the datakey is in the datapacket
        if len(config['datakey']) == 0:
            datakeys = redvypr.data_packets.Datapacket(data).datakeys()
        else:
            if config['datakey'] in data.keys():
                datakeys = [config['datakey']]
            else:
                datakeys = []

        for datakey in datakeys:
            dataparse = data[datakey]
            if (type(dataparse) == bytes):
                try:
                    dataparse = dataparse.decode('UTF-8')
                except:
                    logger.debug(funcname + ' Could not decode:',exc_info=True)
                    continue
            if (type(dataparse) == str) and (len(dataparse) > 2):
                print('Parsing data at datakey {}:{}'.format(datakey,dataparse))

                try:
                    msg = pynmea2.parse(dataparse)
                    talker = msg.talker
                    sentence_type = msg.sentence_type
                    daddr = redvypr.RedvyprAddress(data)
                    devname = sentence_type + '_' + daddr.devicename
                    print('Devicename', daddr.devicename)
                    data_parsed = redvypr.data_packets.create_datadict(data=msg.talker, datakey='talker', device=devname)
                    data_parsed['sentence_type'] = sentence_type
                    #attr = dir(msg)
                    #for a in attr:
                    #    d = getattr(msg,a)
                    #    #print('da',a,d)
                    #    if not(a.startswith('_')):
                    #        if(type(d) == float) or (type(d) == int) or (type(d) == str) or (type(d) == bool):
                    #            data_parsed[a] = d
                    for field in msg.fields:
                        field_short = field[1]
                        field_descr = field[0]
                        data_parsed[field_short] = getattr(msg,field_short)

                    print('Parsing of type {} suceeded:{}'.format(sentence_type,msg))
                    #print('Longitude',msg.longitude,msg.latitude)
                    print('Data parsed',data_parsed)
                    print('----Done-----')
                    dataqueue.put(data_parsed)
                    #if msgtype == 'GGA':
                    #    #tGGA = msg.timestamp
                    #    data_parsed = redvypr.data_packets.datapacket(data = msg.longitude, datakey='lon',device = devname)
                    #    data_parsed['latitude'] = msg.latitude
                    #    dataqueue.put(data_parsed)
                except:
                    logger.debug('NMEA Parsing failed:', exc_info=True)





class initDeviceWidget(QtWidgets.QWidget):
    connect = QtCore.pyqtSignal(
        RedvyprDevice)  # Signal requesting a connect of the datainqueue with available dataoutqueues of other devices
    def __init__(self,device=None):
        super(QtWidgets.QWidget, self).__init__()
        layout        = QtWidgets.QVBoxLayout(self)
        self.device   = device

        self.messagelist = QtWidgets.QListWidget()
        self.sub_button = QtWidgets.QPushButton('Subscribe')
        self.sub_button.clicked.connect(self.subscribe_clicked)
        self.start_button = QtWidgets.QPushButton('Start')
        self.start_button.clicked.connect(self.start_clicked)
        self.start_button.setCheckable(True)
        layout.addWidget(self.messagelist)
        layout.addWidget(self.sub_button)
        layout.addWidget(self.start_button)
        self.config_widgets = []
        self.config_widgets.append(self.sub_button)
        self.config_widgets.append(self.messagelist)

        self.statustimer = QtCore.QTimer()
        self.statustimer.timeout.connect(self.update_buttons)
        self.statustimer.start(500)

        for i,NMEA_message in enumerate(['GGA','VTG']):
            item = QtWidgets.QListWidgetItem(NMEA_message)
            #item.setFlags(QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsEnabled)
            item.setFlags(item.flags() | QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsEnabled)
            item.setCheckState(QtCore.Qt.Checked)
            self.messagelist.addItem(item)


    def start_clicked(self):
        button = self.sender()
        print('start utton',button.isChecked())
        if button.isChecked():
            logger.debug("button pressed")
            button.setText('Starting')
            self.device.thread_start()
            #self.start_button.setChecked(True)
            # self.device_start.emit(self.device)
        else:
            logger.debug('button released')
            # button.setText('Stopping')
            #self.start_button.setChecked(False)
            self.device.thread_stop()

    def subscribe_clicked(self):
        funcname = __name__ + '.subscribe_clicked()'
        print(funcname)
        self.connect.emit(self.device)

    def update_buttons(self):
        """ Updating all buttons depending on the thread status (if its alive, graying out things)
        """

        status = self.device.get_thread_status()
        thread_status = status['thread_running']
        # Running
        if (thread_status):
            self.start_button.setText('Stop')
            self.start_button.setChecked(True)
            for w in self.config_widgets:
                w.setEnabled(False)
        # Not running
        else:
            self.start_button.setText('Start')
            for w in self.config_widgets:
                w.setEnabled(True)

            # Check if an error occured and the startbutton
            if (self.start_button.isChecked()):
                self.start_button.setChecked(False)
            # self.conbtn.setEnabled(True)





"""

zeromq device

Configuration options for a network device:

.. code-block::

- devicemodulename: network_device  
  deviceconfig:
    name: tcpserial1 # The name, must be unique
    config:
       address: <ip> # Address IP, localhost. <broadcast> for UDP broadcast in local network. <ip> for local ip
       port: 10001 # Port
       serialize: str # yaml,str default yaml
       protocol: tcp # tcp, udp default tcp
       direction: publish # publish, receive default receive
       data: nmea # dictionary keys, default all
       tcp_reconnect: True # Try to reconnect to host if connection was closed
       tcp_numreconnect: 10 # The number of reconnection attempts
"""


import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import logging
import sys
import threading
#from apt_pkg import config
import yaml
import copy
import zmq
import socket
from redvypr.device import RedvyprDevice
from redvypr.data_packets import check_for_command



zmq_context = zmq.Context()
description = 'Send or reveives data via a zeromq PUB/SUB socket'

config_template = {}
config_template['name']      = 'zeromq'
config_template['address']   = {'type': 'str','default':'127.0.0.1'}
config_template['port']      = {'type': 'int','default':18196,'range':[0,65535]}
config_template['direction'] = {'type': 'str', 'options': ['receive', 'publish'],'default':'receive'}
config_template['data']      = {'type': 'str','default':'data'}
config_template['serialize'] = {'type': 'str', 'options': ['yaml', 'str'],'default':'yaml'}
config_template['redvypr_device'] = {}
config_template['redvypr_device']['publishes']   = True
config_template['redvypr_device']['subscribes']  = False
config_template['redvypr_device']['description'] = description
redvypr_devicemodule = True

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('zeromq_device')
logger.setLevel(logging.DEBUG)



#https://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib
def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP

def yaml_dump(data):
    #return yaml.dump(data,default_flow_style=False)
    return yaml.dump(data,explicit_end=True,explicit_start=True)

def rawcopy(data):
    """
    Functions that simply returns the input data, used as the basic field conversion function
    """
    return data


def start_send(dataqueue, datainqueue, statusqueue, config=None):
    """ zeromq publishing.
    """
    funcname = __name__ + '.start_send()'
    logger.debug(funcname + ':Starting zeromq send thread')
    npackets     = 0 # Number of packets
    bytes_read   = 0

    pub = zmq_context.socket(zmq.PUB)
    url = 'tcp://' + config['address'] + ':' + str(config['port'])
    logger.debug(funcname + ':Start receiving data from url {:s}'.format(url))
    pub.connect(url)

    # Some variables for status
    tstatus    = time.time()
    dtstatus   = 5 # Send a status message every dtstatus seconds
    npackets   = 0  # Number packets received via the datainqueue
    bytes_sent = 0  # Number packets received via the datainqueue

    if(config['serialize'] == 'str'):
        serialize = str
    elif (config['serialize'] == 'yaml'):
        serialize = yaml_dump #
    elif (config['serialize'] == 'raw'):
        serialize = rawcopy  #
    else:
        serialize = yaml_dump #
            
    while True:
        if True:
            data = datainqueue.get()
            command = check_for_command(data)
            if(command is not None):
                logger.debug('Got a command', command)
                break

        npackets += 1
        # Call the send_data function to create a binary sendable datastream

        #
        if (config['data'][0] == 'all'):
            datab = serialize(data_dict).encode('utf-8')
        else:
            if (type(config['data']) == str):
                config['data'] = [config['data']]

            datab = b''
            for key in config['data']:
                datab += serialize(data_dict[key]).encode('utf-8')
        datab      = send_data(data,config)
        pub.send(datab)
        bytes_sent += len(datab)

        # Sending a status message
        if((time.time() - tstatus) > dtstatus):
            statusdata = {'npackets':npackets}
            statusdata['bytes_sent'] = bytes_sent
            tstatus = time.time()
            try:
                statusqueue.put_nowait(statusdata)
            except: # If the queue is full
                pass


def start_recv(dataqueue, datainqueue, statusqueue, config=None,device_info=None):
    """ zeromq receiving data
    """
    funcname = __name__ + '.start_recv()'

    sub = zmq_context.socket(zmq.SUB)
    url = 'tcp://' + config['address'] + ':' + str(config['port'])
    logger.debug(funcname + ':Start receiving data from url {:s}'.format(url))
    sub.setsockopt(zmq.RCVTIMEO, 200)
    sub.connect(url)
    # subscribe to topic '', TODO, do so
    sub.setsockopt(zmq.SUBSCRIBE, b'')

    datapackets = 0
    bytes_read  = 0

    # Some variables for status
    tstatus = time.time()
    try:
        dtstatus = config['dtstatus']
    except:
        dtstatus = 2 # Send a status message every dtstatus seconds
        
    #
    npackets = 0 # Number packets received
    while True:
        try:
            data = datainqueue.get(block=False)
        except:
            data = None

        if(data is not None):
            command = check_for_command(data,thread_uuid=device_info['thread_uuid'])
            logger.debug('Got a command {:s}'.format(str(data)))
            if (command is not None):
                logger.debug(funcname + ': received command:' + str(command) + ', stopping now')
                sub.close()
                logger.debug(funcname + ': zeromq port closed.')
                return

        try:
            #datab = sub.recv(zmq.NOBLOCK)
            datab = sub.recv()
            #print('Got data',datab)
        except Exception as e:
            #logger.debug(funcname + ':' + str(e))
            datab = b''


        if(len(datab)>0):
            t = time.time()
            bytes_read += len(datab)
            # Check what data we are expecting and convert it accordingly
            if(config['serialize'] == 'yaml'):
                for databs in datab.split(b'...\n'): # Split the text into single subpackets
                    try:
                        data = yaml.safe_load(databs)
                        #print(datab)
                        #print(data)
                    except Exception as e:
                        logger.debug(funcname + ': Could not decode message {:s}'.format(str(datab)))
                        logger.debug(funcname + ': Could not decode message  with supposed format {:s} into something useful.'.format(str(config['data'])))
                        data = None

                    if((data is not None) and (type(data) == dict)):
                        dataqueue.put(data)
                        datapackets += 1

            else: # Put the "str" data into the packet with the key in "data"
                key = config['data']
                data = datab.decode('utf-8')
                datan = {'t':t}
                datan[key] = data
                try:
                    dataqueue.put(datan)
                except Exception as e:
                    print('Exception',e)

                datapackets += 1



        # Sending a status message
        if((time.time() - tstatus) > dtstatus):
            statusdata = {}
            statusdata['bytes_read']  = bytes_read
            statusdata['datapackets'] = datapackets
            statusdata['time'] = str(datetime.datetime.now())
            tstatus = time.time()
            #statusdata['clients'] = []
            statusdata['config'] = copy.deepcopy(config)
                
            try:
                statusqueue.put_nowait(statusdata)
            except: # If the queue is full
                pass

def start(device_info, config=None, dataqueue=None, datainqueue=None, statusqueue=None):
    funcname = __name__ + '.start():'
    logger.debug(funcname)
    if (config['direction'] == 'publish'):
        logger.info(__name__ + ':Start to serve data on address:' + str(config))
        # start_send(self.dataqueue,self.datainqueue,self.statusqueue,config=self.config)
    elif (config['direction'] == 'receive'):
        logger.info('Start to receive data from address:' + str(config))
        start_recv(dataqueue, datainqueue, statusqueue, config=config, device_info=device_info)
        






import datetime
import numpy as np
import logging
import sys
import threading
import copy
import yaml
import json
import typing
import pydantic
from PyQt5 import QtWidgets, QtCore, QtGui

import redvypr
from redvypr.device import RedvyprDevice, device_start_standard
from redvypr.data_packets import check_for_command
#from redvypr.packet_statistics import get_keys_from_data
#import redvypr.packet_statistic as redvypr_packet_statistic
import redvypr.data_packets as data_packets
import redvypr.gui as gui
#import redvypr.config as redvypr_config
from redvypr.redvypr_address import RedvyprAddress
#from redvypr.devices.plot import plot_widgets
from redvypr.devices.plot import XYplotWidget
import redvypr.files as redvypr_files

_icon_file = redvypr_files.icon_file

redvypr_devicemodule = True

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('xyplot')
logger.setLevel(logging.INFO)


class DeviceBaseConfig(pydantic.BaseModel):
    publishes: bool = False
    subscribes: bool = True
    description: str = 'Device to plot XY-Data'
    gui_icon: str = 'ph.chart-line-fill'




class DeviceCustomConfig(XYplotWidget.configXYplot):
    pass


# Use the standard start function as the start function
start = device_start_standard

class displayDeviceWidget(QtWidgets.QWidget):
    def __init__(self,device=None,tabwidget=None):
        funcname = __name__ + '__init__():'
        logger.debug(funcname)
        super(QtWidgets.QWidget, self).__init__()
        self.device = device
        self.layout = QtWidgets.QGridLayout(self)
        self.xyplot = XYplotWidget.XYPlotWidget(redvypr_device=device, config=self.device.custom_config)
        self.layout.addWidget(self.xyplot)
        self.device.config_changed_signal.connect(self.config_changed)

    def config_changed(self):
        print('XYplot config changed')
        print('Config',self.device.custom_config)
        # Check if subscriptions need to be changed
        self.xyplot.config = self.device.custom_config
        self.xyplot.apply_config()

    def update_data(self, data):
        funcname = __name__ + '.update_data():'
        #print(funcname)
        self.xyplot.update_plot(data)
